-- 65xx compatible microprocessor core
--
-- Version : 1.0 Single clock
--
--

library ieee;
  use ieee.std_logic_1164.all;
  use ieee.numeric_std.all ;


entity RB65 is
  Port (
    A		: out	std_logic_vector(15 downto 0);
    Din		: in	std_logic_vector(7 downto 0);
    Dout	: out 	std_logic_vector(7 downto 0);
    IRQ		: in	std_logic;
    NMI		: in	std_logic;
    PHI0	: in	std_logic;
    RDY		: in	std_logic;
    RESET	: in	std_logic;
    RW		: out	std_logic;		-- Read/Write line
    SO_n	: in	std_logic;
    SYNC	: out	std_logic;

-- Debug purposes
    ALU_a2	: out 	std_logic_vector(7 downto 0);
    ALU_b2	: out 	std_logic_vector(7 downto 0);
    ALU_r2	: out 	std_logic_vector(7 downto 0);
    ALU_C2	: out	std_logic;    
    ALU_V2	: out	std_logic;    
    OpcC	: out	std_logic_vector(4 downto 0);	-- Opcode step counter
    OpcCR	: out	std_logic;
    xx1,
    Version	: out 	std_logic_vector(7 downto 0)
  );
end RB65;


architecture AR_RB65 of RB65 is
  signal Accu   : unsigned(7 downto 0) := x"FF";	-- 16 bits reg. A
  signal PC     : unsigned(15 downto 0); 		-- Program Counter
  signal SP     : unsigned(7 downto 0) := x"FF";	-- 16 bits reg. Stack Pointer
  signal X     	: unsigned(7 downto 0) := x"FF";	-- 16 bits reg. X
  signal Y      : unsigned(7 downto 0) := x"FF";	-- 16 bits reg. Y

  signal P      : unsigned(7 downto 0) := x"FF";	-- 8 bits Status reg.
  signal PCh    : unsigned(7 downto 0) := x"FF";	-- pre Program Counter High
  signal PCl    : unsigned(7 downto 0) := x"FF";	-- pre Program Counter Low

  signal Tmp    : unsigned(15 downto 0);		-- Temporary buffer


--  ALU  --
  signal ALU_a,						-- input 1 (Accu)
         ALU_b,
         ALU_r		: unsigned(7 downto 0) := x"00";	-- result

  signal ALU_C,						-- output ALU Carry
         ALU_Ci,					-- input ALU Carry
         ALU_D,						-- input ALU Decimal
         ALU_V,						-- output ALU Overflow
         ALU_Z,						-- output ALU Zero
         ALU_N		: std_logic := '0';		-- output ALU Negative

  signal NibbleL,
         NibbleL2,
         NibbleH,
	 NibbleH2	: unsigned(4 downto 0) := "00000";	

  signal Operation	: unsigned(4 downto 0);		-- ALU operation


  signal AddressMode	: unsigned(1 downto 0) := "10";
  signal D_out		: unsigned(7 downto 0) := x"FF";
  signal IntData    	: unsigned(7 downto 0);		-- Data latch
  signal Opcode    	: unsigned(7 downto 0);		-- Opcode

  signal Branch 	: std_logic := '0';
  signal IRQ_in		: std_logic := '0';
  signal NMI_busy	: std_logic := '0';
  signal NMI_handled	: std_logic := '1';
  signal NMI_in		: std_logic := '0';
  signal OpcC_reset	: std_logic := '0';
  signal Reset_in	: std_logic := '0';		-- RESET detected
  signal RW_i		: std_logic;			-- 0 = write, 1 = read


  constant amPC		: unsigned(1 downto 0) := "00";
  constant amSP		: unsigned(1 downto 0) := "01";
  constant amTmp	: unsigned(1 downto 0) := "10";
  constant amPhl	: unsigned(1 downto 0) := "11";


  constant Flag_C 	: integer := 0;
  constant Flag_Z 	: integer := 1;
  constant Flag_I 	: integer := 2;
  constant Flag_D 	: integer := 3;
  constant Flag_B 	: integer := 4;
  constant Flag_1 	: integer := 5;
  constant Flag_V 	: integer := 6;
  constant Flag_N 	: integer := 7;

  constant opAND 	: unsigned(4 downto 0) := "00000";
  constant opOR  	: unsigned(4 downto 0) := "00001";
  constant opEOR 	: unsigned(4 downto 0) := "00010";
  constant opASL 	: unsigned(4 downto 0) := "00011";
  constant opLSR 	: unsigned(4 downto 0) := "00100";
  constant opROL 	: unsigned(4 downto 0) := "00101";
  constant opROR 	: unsigned(4 downto 0) := "00110";
  constant opINC 	: unsigned(4 downto 0) := "00111";
  constant opDEC 	: unsigned(4 downto 0) := "01000";
  constant opADC 	: unsigned(4 downto 0) := "01001";
  constant opCMP 	: unsigned(4 downto 0) := "01010";
  constant opBIT 	: unsigned(4 downto 0) := "01011";
  constant opSBC	: unsigned(4 downto 0) := "01100";
  constant opBRA 	: unsigned(4 downto 0) := "01101";
  constant opLDx 	: unsigned(4 downto 0) := "01110";
  constant opTxx 	: unsigned(4 downto 0) := "01111";
  constant opASLa 	: unsigned(4 downto 0) := "10011";
  constant opLSRa 	: unsigned(4 downto 0) := "10100";
  constant opROLa 	: unsigned(4 downto 0) := "10101";
  constant opRORa 	: unsigned(4 downto 0) := "10110";

			
begin
  Version <= x"11" when (Reset = '0') else std_logic_vector(P);
  OpcCR <= OpcC_Reset;

  Dout <= std_logic_vector(D_out);					
  IntData <= unsigned(Din) when (PHI0 = '0');		-- read data
  Opcode <= unsigned(Din) when ((PHI0 = '0') and (OpcC_Reset = '1'));
  RW <= RW_i;
  SYNC <= '1' when ((PHI0 = '0') and (OpcC_Reset = '1')) else '0';


  with AddressMode select A <=
    std_logic_vector(PC)		when amPC,
    x"01" & std_logic_vector(SP)	when amSP,
    std_logic_vector(Tmp)		when amTmp,
    std_logic_vector(PCh & PCl)		when others;


ALU_a2 <= std_logic_vector(ALU_a);
ALU_b2 <= std_logic_vector(ALU_b);
ALU_r2 <= std_logic_vector(ALU_r);
ALU_C2 <= std_logic(ALU_C);
ALU_V2 <= std_logic(ALU_V);
xx1 <= std_logic_vector(Accu(7 downto 0));


  process(NMI)
  begin
    if ((NMI = '1') and (NMI_in = '1') and (NMI_handled = '1')) then
      NMI_in <= '0';
    elsif (falling_edge(NMI) and (NMI_in = '0')) then
      NMI_in <= '1';
    end if;
  end process;


  process(RESET, PHI0)
    variable OpcCounter : unsigned(4 downto 0) := "00000";	-- Opcode step counter, PHI2 is LSB
  begin
    if (RESET = '0') then
      AddressMode <= amTmp;
      NMI_busy <= '0';
      NMI_handled <= '1';
      OpcC <= std_logic_vector(OpcCounter);
      OpcC_reset <= '1';
      Operation <= "11111";
      PC <= x"FFFC";
      Reset_in <= '1';
      RW_i <= '1';
      Tmp <= x"FFFC";
    elsif rising_edge(PHI0) then

--  Handle SO input
      if (SO_n = '0') then
        P(Flag_V) <= '1';
      end if;

--  Check if RDY input is negated
      if (RDY = '1') then
      
--  Handle possible NMI/IRQ at start of every instruction
        if (OpcC_Reset = '1') then
          OpcCounter := "00010";
          OpcC_Reset <= '0';

          if (   (NMI_in = '1')
              or ((IRQ = '0') and (P(Flag_I) = '0'))
             ) then			-- push HB PC to Stack
            AddressMode <= amSP;
            RW_i <= '0';
            D_out <= PC(15 downto 8);

            if (NMI_in = '1') then
              NMI_busy <= '1';
              NMI_handled <= '0';
            else
             IRQ_in <= '1';
            end if;
          else
            AddressMode <= amPC;
          end if;
        else
          OpcCounter := OpcCounter + 2;		-- no check on overflow needed
        end if;

        OpcC <= std_logic_vector(OpcCounter);

--  === Reset ===
        if (Reset_in = '1') then
          case OpcCounter is
            when "00010" =>
              AddressMode <= amTmp;
              Tmp <= x"FFFC";
              IRQ_in <= '0';
              RW_i <= '1';
            when "00100" =>
              PC(7 downto 0) <= IntData;	-- read low byte reset routine
              Tmp <= x"FFFD";
            when "00110" =>
              AddressMode <= amPC; 	-- output read vector
              PC(15 downto 8) <= IntData;	-- read high byte reset routine
              NMI_busy <= '0';
              NMI_handled <= '1';
              OpcC_reset <= '1';
              Reset_in <= '0';
            when others => null;
          end case;
  

--  === IRQ / NMI ===
        elsif ((NMI_busy = '1') or (IRQ_in = '1')) then
          case OpcCounter is
            when "00100" =>			-- push LB PC to Stack
              SP <= SP - 1;
              D_out <= PC(7 downto 0);
              RW_i <= '0';

              if (NMI_busy = '0') then
                P(Flag_B) <= '0';
              end if;
            when "00110" =>			-- push Status Register to Stack
              SP <= SP - 1;
              D_out <= P;
              RW_i <= '0';
            when "01000" =>
              AddressMode <= amTmp;
              P(Flag_I) <= '0';
              RW_i <= '1';

              if (NMI_busy = '1') then
                Tmp <= x"FFFA";
              else
                Tmp <= x"FFFE";
              end if;
            when "01010" =>
              PC(7 downto 0) <= IntData;	-- read low byte reset routine
              if (NMI_busy = '1') then
                Tmp <= x"FFFB";
              else
                Tmp <= x"FFFF";
              end if;
            when "01100" =>
              AddressMode <= amPC;	 	-- output read vector
              PC(15 downto 8) <= IntData;	-- read high byte reset routine
            when "01110" =>		-- COMPATIBILITY
              if (IRQ_in = '1') then
                IRQ_in <= '0';
              else
                NMI_busy <= '0';
                NMI_handled <= '1';
              end if;
              OpcC_reset <= '1';
            when others => null;
          end case;


--  === All opcodes ===
        else
          case Opcode is

            when x"00" =>		-- BRK
              case OpcCounter is
                when "00010" =>			-- increment PC twice
                  PC <= PC + 2;
                  AddressMode <= amSP;
                  RW_i <= '0';
                  D_out <= PC(15 downto 8);	-- push HB PC to Stack
                when "00100" =>			-- push LB PC to Stack
                  SP <= SP - 1;
                  D_out <= PC(7 downto 0);
                when "00110" =>			-- push Status Register to Stack
                  SP <= SP - 1;
                  D_out <= P;
                when "01000" =>
                  AddressMode <= amTmp;
                  Tmp <= x"FFFE";
                  P(Flag_B) <= '1';
                  RW_i <= '1';
                when "01010" =>
                  PC(7 downto 0) <= IntData;	-- read low byte BRK routine
                  Tmp <= x"FFFF";
                when "01100" =>
                  AddressMode <= amPC;		-- store read vector into PC
                  PC(15 downto 8) <= IntData;	-- read high byte BRK routine
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"01" |		-- ORA ($xx,X)
                 x"21" |		-- AND ($xx,X)
                 x"41" |		-- EOR ($xx,X)
                 x"61" |		-- ADC ($xx,X)
                 x"81" |		-- STA ($xx,X)
                 x"A1" |		-- LDA ($xx,X)
                 x"C1" |		-- CMP ($xx,X)
                 x"E1" =>		-- SBC ($xx,X)
              case OpcCounter is
                when "00010" =>			-- read zero page address
                  PC <= PC + 1;

                  case Opcode(7 downto 5) is
                    when "000" => Operation <= opOR;
                    when "001" => Operation <= opAND;
                    when "010" => Operation <= opEOR;
                    when "011" => Operation <= opADC;
                    when "101" => Operation <= opLDx;
                    when "110" => Operation <= opCMP;
                    when "111" => Operation <= opSBC;
                    when others => null;
                  end case;
                when "00100" =>			-- read LB indexed address
                  AddressMode <= amTmp;
                  Tmp <= x"00" & (IntData + X(7 downto 0));
                when "00110" =>			-- read HB indexed address
                  Tmp <= (Tmp + 1) and x"00FF";
                  PCl <= IntData;		-- store LB indexed address
                when "01000" =>			-- read byte at indexed address
                  ALU_a <= Accu(7 downto 0);
                  AddressMode <= amPhl;
                  PCh <= IntData;		-- store HB indexed address
                  
                  if (Opcode(7 downto 5) = "100") then
                    D_out <= Accu(7 downto 0);
                    RW_i <= '0';
                  end if;
                when "01010" =>			-- calculate and store results
                  AddressMode <= amPC;
                  PC <= PC + 1;

                  if (Opcode(7 downto 5) = "100") then
                    RW_i <= '1';
                  else
                    P(Flag_N) <= ALU_N;
                    P(Flag_Z) <= ALU_Z;

                    case Opcode(6 downto 5) is
                      when "11" => P(Flag_V) <= ALU_V;
                      when others => null;
                    end case;

                    case Opcode(7 downto 5) is
                      when "011" | "110" | "111" => P(Flag_C) <= ALU_C;
                      when others => null;
                    end case;

                    if (    (Opcode(7 downto 5) /= "100")
		        and (Opcode(7 downto 5) /= "110")) then
                      Accu(7 downto 0) <= ALU_r;
                    end if;
                  end if;
                when "01100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"05" |		-- ORA $xx
                 x"25" |		-- AND $xx
                 x"45" |		-- EOR $xx
                 x"65" |		-- ADC $xx
                 x"A4" |		-- LDY $xx
                 x"A5" |		-- LDA $xx
                 x"A6" |		-- LDX $xx
                 x"E5" |		-- SBC $xx

                 x"15" |		-- ORA $xx,X
                 x"35" |		-- AND $xx,X
                 x"55" |		-- EOR $xx,X
                 x"75" |		-- ADC $xx,X
                 x"B4" |		-- LDY $xx,X
                 x"B5" |		-- LDA $xx,X
                 x"B6" |		-- LDX $xx,Y
                 x"F5" =>		-- SBC $xx,X
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;
                  ALU_a <= Accu(7 downto 0);

                  case Opcode(7 downto 5) is
                    when "000" => Operation <= opOR;
                    when "001" => Operation <= opAND;
                    when "010" => Operation <= opEOR;
                    when "011" => Operation <= opADC;
                    when "101" => Operation <= opLDx;
                    when "111" => Operation <= opSBC;
                    when others => null;
                  end case;
                when "00100" =>		-- read zero page address
                  AddressMode <= amTmp;

                  if (Opcode(4) = '0') then
                    Tmp <= x"00" & IntData;
                  elsif (Opcode(1) = '0') then  
                    Tmp <= x"00" & (IntData + X(7 downto 0));
                  else 
                    Tmp <= x"00" & (IntData + Y(7 downto 0));
                  end if;
                when "00110" =>		-- calculate and store results
                  AddressMode <= amPC;
                  PC <= PC + 1;
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;

                  if (Opcode(6 downto 5) = "11") then
                    P(Flag_V) <= ALU_V;
                    P(Flag_C) <= ALU_C;
                  end if;

                  case Opcode(1 downto 0) is
                    when "00" => Y(7 downto 0) <= ALU_r;
                    when "10" => X(7 downto 0) <= ALU_r;
                    when others => Accu(7 downto 0) <= ALU_r;
                  end case;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"06" |		-- ASL $xx
                 x"26" |		-- ROL $xx
                 x"46" |		-- LSR $xx
                 x"66" |		-- ROR $xx
                 x"C6" |		-- DEC $xx
                 x"E6" | 		-- INC $xx

                 x"16" |		-- ASL $xx,X
                 x"36" |		-- ROL $xx,X
                 x"56" |		-- LSR $xx,X
                 x"76" |		-- ROR $xx,X
                 x"D6" |		-- DEC $xx,X
                 x"F6" =>		-- INC $xx,X
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;
                when "00100" =>		-- read LB indexed address
                  AddressMode <= amTmp;

                  if (Opcode(4) = '0') then
                    Tmp <= x"00" & IntData;
                  else 
                    Tmp <= x"00" & (IntData + X(7 downto 0));
                  end if;

                  case Opcode(7 downto 5) is
                    when "000" => Operation <= opASL;
                    when "001" => Operation <= opROL;
                    when "010" => Operation <= opLSR;
                    when "011" => Operation <= opROR;
                    when "110" => Operation <= opDEC;
                    when "111" => Operation <= opINC;
                    when others => null;
                  end case;
                when "00110" =>		-- calculate and store results
                  RW_i <= '0';
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;
                  D_out <= ALU_r;

                  if (Opcode(7) = '0') then
                    P(Flag_C) <= ALU_C;
                  end if;
                when "01000" =>
                  AddressMode <= amPC;
                  RW_i <= '1';
                  PC <= PC + 1;
                when "01010" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"08" |		-- PHP
                 x"48" =>		-- PHA
              case OpcCounter is
                when "00010" =>		-- increment PC / push to stack
                  AddressMode <= amSP;
                  PC <= PC + 1;
                  RW_i <= '0';

                  if (Opcode(6) = '0') then
                    D_out <= P or x"10";
                  else  
                    D_out <= Accu(7 downto 0);
                  end if;
                when "00100" =>
                  AddressMode <= amPC;
                  SP <= SP - 1;
                  RW_i <= '1';
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"09" |		-- ORA #n
                 x"29" |		-- AND #n
                 x"49" |		-- EOR #n
                 x"69" |		-- ADC #n
                 x"E9" =>		-- SBC #n
              case OpcCounter is
                when "00010" =>		-- read byte
                  ALU_a <= Accu(7 downto 0);
                  PC <= PC + 1;

                  case Opcode(7 downto 5) is
                    when "000" => Operation <= opOR;
                    when "001" => Operation <= opAND;
                    when "010" => Operation <= opEOR;
                    when "011" => Operation <= opADC;
                    when "111" => Operation <= opSBC;
                    when others => null;
                  end case;
                when "00100" =>		-- calculate and store results
                  if (Opcode(6 downto 5) = "11") then
                    P(Flag_V) <= ALU_V;
                    P(Flag_C) <= ALU_C;
                  end if;

                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;
                  Accu(7 downto 0) <= ALU_r;
                  PC <= PC + 1;
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"0A" |		-- ASL A
                 x"2A" |		-- ROL A
                 x"4A" |		-- LSR A
                 x"6A" =>		-- ROR A
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  ALU_a <= Accu(7 downto 0);
                  PC <= PC + 1;

                  case Opcode(6 downto 5) is
                    when "00" => Operation <= opASLa;
                    when "01" => Operation <= opROLa;
                    when "10" => Operation <= opLSRa;
                    when "11" => Operation <= opRORa;
                    when others => null;
                  end case;
                when "00100" =>		-- calculate and store results
                  Accu(7 downto 0) <= ALU_r;
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;
                  P(Flag_C) <= ALU_C;
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"0D" |		-- ORA $xxxx
                 x"2D" |		-- AND $xxxx
                 x"4D" |		-- EOR $xxxx
                 x"6D" |		-- ADC $xxxx
                 x"AC" |		-- LDY $xxxx
                 x"AD" |		-- LDA $xxxx
                 x"AE" |		-- LDX $xxxx
                 x"ED" |		-- SBC $xxxx

                 x"1D" |		-- ORA $xxxx,X
                 x"3D" |		-- AND $xxxx,X
                 x"5D" |		-- EOR $xxxx,X
                 x"7D" |		-- ADC $xxxx,X
                 x"FD" |		-- SBC $xxxx,X
                 x"BC" |		-- LDY $xxxx,X
                 x"BD" |		-- LDA $xxxx,X

                 x"B9" |		-- LDA $xxxx,Y
                 x"BE" |		-- LDX $xxxx,Y
                 x"19" |		-- ORA $xxxx,Y
                 x"39" |		-- AND $xxxx,Y
                 x"59" |		-- EOR $xxxx,Y
                 x"79" |		-- ADC $xxxx,Y
                 x"F9" =>		-- SBC $xxxx,Y
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;

                  case Opcode(7 downto 5) is
                    when "000" => Operation <= opOR;
                    when "001" => Operation <= opAND;
                    when "010" => Operation <= opEOR;
                    when "011" => Operation <= opADC;
                    when "101" => Operation <= opLDx;
                    when "111" => Operation <= opSBC;
                    when others => null;
                  end case;
                when "00100" =>		-- read LB address
                  PCl <= IntData;
                  PC <= PC + 1;
                  
                when "00110" =>		-- read HB address
                  ALU_a <= Accu(7 downto 0);
                  AddressMode <= amTmp;

                  if (Opcode(4) = '0') then
                    Tmp <= IntData & PCl;
		  elsif (Opcode(2 downto 1) = "10") then
                    Tmp <= (IntData & PCl) + X;
		  else
                    Tmp <= (IntData & PCl) + Y;
                  end if;
                when "01000" =>		-- calculate and store results
                  AddressMode <= amPC;
                  PC <= PC + 1;
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;

                  case Opcode(6 downto 5) is
                    when "11" =>
                      P(Flag_V) <= ALU_V;
                      P(Flag_C) <= ALU_C;
                    when others => null;
                  end case;

                  case Opcode(1 downto 0) is
                    when "00" => Y(7 downto 0) <= ALU_r;
                    when "10" => X(7 downto 0) <= ALU_r;
                    when others => Accu(7 downto 0) <= ALU_r;
                  end case;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"0E" |		-- ASL $xxxx
                 x"2E" |		-- ROL $xxxx
                 x"4E" |		-- LSR $xxxx
                 x"6E" |		-- ROR $xxxx
                 x"CE" |		-- DEC $xxxx
                 x"EE" |		-- INC $xxxx

                 x"1E" |		-- ASL $xxxx,X
                 x"3E" |		-- ROL $xxxx,X
                 x"5E" |		-- LSR $xxxx,X
                 x"7E" |		-- ROR $xxxx,X
                 x"DE" |		-- DEC $xxxx,X
                 x"FE" =>		-- INC $xxxx,X
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;
                when "00100" =>		-- read LB address
                  PC <= PC + 1;
                  PCl <= IntData;
                when "00110" =>		-- read HB address
                  AddressMode <= amTmp;

                  if (Opcode(4) = '0') then
                    Tmp <= IntData & PCl;
		  else
                    Tmp <= (IntData & PCl) + X;
                  end if;

                  case Opcode(7 downto 5) is
                    when "000" => Operation <= opASL;
                    when "001" => Operation <= opROL;
                    when "010" => Operation <= opLSR;
                    when "011" => Operation <= opROR;
                    when "110" => Operation <= opDEC;
                    when "111" => Operation <= opINC;
                    when others => null;
                  end case;
                when "01000" =>		-- calculate and store results
                  RW_i <= '0';
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;
                  D_out <= ALU_r;

                  if (Opcode(7) = '0') then
                    P(Flag_C) <= ALU_C;
                  end if;
                when "01010" =>
                  AddressMode <= amPC;
                  PC <= PC + 1;
                  RW_i <= '1';
                when "01100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"10" |		-- BPL
                 x"30" |		-- BMI
                 x"50" |		-- BVC
                 x"70" |		-- BVS
                 x"90" |		-- BCC
                 x"B0" |		-- BCS
                 x"D0" |		-- BNE
                 x"F0" =>		-- BEQ
              case OpcCounter is
                when "00010" =>			-- read offset
                  PC <= PC + 1;
                  Operation <= opBRA;
                when "00100" =>			-- decide if branch is needed
                  ALU_a <= IntData;		-- store for ALU
                  PC <= PC + 1;			-- increase PC 

                  if (Branch = '0') then
                    OpcC_reset <= '1';		-- no branch taken
                  end if;
                when "00110" =>			-- branch taken
                  PC(7 downto 0) <= ALU_r;

    	-- page boundary crossed
                  if ((ALU_a(7) = '0') and (ALU_C = '1'))then
                    PC(15 downto 8) <= PC(15 downto 8) + 1;
                  elsif ((ALU_a(7) = '1') and (ALU_C = '0'))then
                    PC(15 downto 8) <= PC(15 downto 8) - 1;
                  else  
                    OpcC_reset <= '1';	-- no page boundary crossed
                  end if;
                when "01000" =>		-- page boundary crossd COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"11" |		-- ORA ($xx),Y
                 x"31" |		-- AND ($xx),Y
                 x"51" |		-- EOR ($xx),Y
                 x"71" |		-- ADC ($xx),Y
                 x"B1" |		-- LDA ($xx),Y
                 x"D1" |		-- CMP ($xx),Y
                 x"F1" =>		-- SBC ($xx),Y
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;

                  case Opcode(7 downto 5) is
                    when "000" => Operation <= opOR;
                    when "001" => Operation <= opAND;
                    when "010" => Operation <= opEOR;
                    when "011" => Operation <= opADC;
                    when "110" => Operation <= opCMP;
                    when "111" => Operation <= opSBC;
                    when others => null;
                  end case;
                when "00100" =>			-- read zero page address
                  AddressMode <= amTmp;
                  ALU_a <= Accu(7 downto 0);
                  Tmp <= x"00" & IntData;
                when "00110" =>			-- read HB indexed address
                  PCl <= IntData;		-- store LB indexed address
                  Tmp <= (Tmp + 1) and x"00FF";
                when "01000" =>			-- read byte at indexed address
                  Tmp <= (IntData  & PCl) + Y;
                  ALU_a <= Accu(7 downto 0);
                when "01010" =>			-- calculate and store results
                  AddressMode <= amPC;
                  PC <= PC + 1;
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;

                  case Opcode(6 downto 5) is
                    when "11" => P(Flag_V) <= ALU_V;
                    when others => null;
                  end case;

                  case Opcode(7 downto 5) is
                    when "011" | "110" | "111" => P(Flag_C) <= ALU_C;
                    when others => null;
                  end case;

                  if (Opcode(7 downto 5) /= "110") then
                    Accu(7 downto 0) <= ALU_r;
    	          end if;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"18" |		-- CLC
                 x"38" |		-- SEC
                 x"58" |		-- CLI
                 x"78" |		-- SEI
                 x"B8" |		-- CLV
                 x"D8" |		-- CLD
                 x"F8" =>		-- SED
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;

                  case Opcode(7 downto 5) is
                    when "000" => P(Flag_C) <= '0';
                    when "001" => P(Flag_C) <= '1';
                    when "010" => P(Flag_I) <= '0';
                    when "011" => P(Flag_I) <= '1';
                    when "101" => P(Flag_V) <= '0';
                    when "110" => P(Flag_D) <= '0';
                    when "111" => P(Flag_D) <= '1';
                    when others => null;
                  end case;
                when "00100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"20" =>		-- JSR
              case OpcCounter is
                when "00010" =>			-- read new address
                  PC <= PC + 1;
                when "00100" =>
                  PC <= PC + 1;
                  PCl <= IntData;		-- read LB address
                when "00110" =>			-- push HB PC to Stack
                  AddressMode <= amSP;
                  D_out <= PC(15 downto 8);
                  PCh <= IntData;		-- read HB address
                  RW_i <= '0';
                when "01000" =>			-- push LB PC to Stack
                  SP <= SP - 1;
                  RW_i <= '0';
                  D_out <= PC(7 downto 0);
                when "01010" =>
                  SP <= SP - 1;
                  AddressMode <= amPC;
                  PC <= PCh & PCl;
                  RW_i <= '1';
                when "01100" =>			-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"24" =>		-- BIT $xx
              case OpcCounter is
                when "00010" =>			-- read zero page address
                  PC <= PC + 1;
                when "00100" =>			-- read zero page address
                  AddressMode <= amTmp;
                  Tmp <= x"00" & IntData;
                  Operation <= opBIT;
                  ALU_a <= Accu(7 downto 0);
                when "00110" =>			-- calculate and store results
                  AddressMode <= amPC;
                  PC <= PC + 1;
                  P(Flag_N) <= ALU_N;
                  P(Flag_V) <= ALU_V;
                  P(Flag_Z) <= ALU_Z;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"28" |		-- PLP
                 x"68" =>		-- PLA
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  SP <= SP + 1;
                when "00100" =>		-- COMPATIBILITY
                  AddressMode <= amSP;
                  Operation <= opTxx;
                when "00110" =>
                  ALU_a <= Intdata;

                  if (Opcode(6) = '0') then
    	-- store read byte into Status register but set bit 5 !!!
                    P <= IntData or x"20";
                  else  
                    Accu(7 downto 0) <= IntData;
          	  end if;
                when "01000" =>
                  AddressMode <= amPC;
                  SP <= SP - 1;

		  if not (Opcode = x"28") then
                    P(Flag_N) <= ALU_N;
                    P(Flag_Z) <= ALU_Z;
                  end if;  

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"2C" =>		-- BIT $xxxx
              case OpcCounter is
                when "00010" =>			-- read zero page address
                  PC <= PC + 1;
                when "00100" =>			-- read LB address
                  PC <= PC + 1;
                  PCl <= IntData;
                when "00110" =>			-- read HB address
                  AddressMode <= amTmp;
                  Tmp <= IntData & PCl;
                  Operation <= opBIT;
                  ALU_a <= Accu(7 downto 0);
                when "01000" =>			-- calculate and store results
                  AddressMode <= amPC;
                  PC <= PC + 1;
                  P(Flag_N) <= ALU_N;
                  P(Flag_V) <= ALU_V;
                  P(Flag_Z) <= ALU_Z;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"40" =>		-- RTI
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  SP <= SP + 1;
                when "00100" =>		-- COMPATIBILITY
                  AddressMode <= amSP;
                when "00110" =>
                  SP <= SP + 1;
                  P <= IntData;			-- store read byte into Status register
                when "01000" =>
                  SP <= SP + 1;
                  PC(7 downto 0) <= IntData;	-- store read byte into PC
                when "01010" =>
                  AddressMode <= amPC;
                  SP <= SP + 1;
                  PC(15 downto 8) <= IntData;	-- store read byte into PC
                when "01100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"4C" =>		-- JMP $xxxx
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                when "00100" =>
                  PC <= PC + 1;
                  PCl <= IntData;	-- read low byte new address
                when "00110" =>
                  PC <= IntData & PCl;	-- read high byte new address
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"60" =>		-- RTS
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  SP <= SP + 1;
                when "00100" =>		-- COMPATIBILITY
                  AddressMode <= amSP;
                when "00110" =>
                  PC(7 downto 0) <= IntData;	-- store read byte into PC
                  SP <= SP + 1;
                when "01000" =>
                  AddressMode <= amPC;
                  PC(15 downto 8) <= IntData;	-- store read byte into PC
                when "01100" =>
                  PC <= PC + 1;
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"6C" =>		-- JMP ($xxxx)
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                when "00100" =>
                  PC <= PC + 1;
                  Tmp(7 downto 0) <= IntData;	-- read low byte indexed address
                when "00110" =>
                  AddressMode <= amTmp;
                  Tmp(15 downto 8) <= IntData;	-- read high byte indexed address
                when "01000" =>
                  Tmp <= Tmp+ 1;
                  PC(7 downto 0) <= IntData;	-- read low byte new address
                when "01010" =>
                  AddressMode <= amPC;
                  PC(15 downto 8) <= IntData;	-- read high byte new address
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"84" |		-- STY $xx
                 x"85" |		-- STA $xx
                 x"86" | 		-- STX $xx

                 x"94" |		-- STY $xx,X
                 x"95" |		-- STA $xx,X
                 x"96" =>		-- STX $xx,Y
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;
                when "00100" =>		-- read LB indexed address
                  AddressMode <= amTmp;
                  RW_i <= '0';

                  if (Opcode(4) = '0') then
                    Tmp <= x"00" & IntData;
                  elsif (Opcode(1) = '1') then
                    Tmp <= x"00" & (IntData + Y(7 downto 0));
                  else
                    Tmp <= x"00" & (IntData + X(7 downto 0));
                  end if;

                  case Opcode(1 downto 0) is
                    when "00" => D_out <= Y(7 downto 0);
                    when "01" => D_out <= Accu(7 downto 0);
                    when "10" => D_out <= X(7 downto 0);
                    when others => null;
                  end case;
                when "00110" =>
                  AddressMode <= amPC;
                  PC <= PC + 1;
                  RW_i <= '1';

                  if (Opcode(4) = '0') then
                    OpcC_reset <= '1';
                  end if;
                when "01000" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"88" |		-- DEY
                 x"C8" |		-- INY
                 x"CA" |		-- DEX
                 x"E8" =>		-- INX
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  Operation <= opTxx;

                  case Opcode is
                    when x"88" | x"C8" =>
                      if (Opcode(6) = '0') then
                        Y <= Y - 1;
                        ALU_a <= Y(7 downto 0) - 1;
                      else
                        Y <= Y + 1;
                        ALU_a <= Y(7 downto 0) + 1;
                      end if;
                    when x"CA" | x"E8" =>
                      if (Opcode(5) = '0') then
                        X <= X - 1;
                        ALU_a <= X(7 downto 0) - 1;
                      else
                        X <= X + 1;
                        ALU_a <= X(7 downto 0) + 1;
                      end if;
                    when others => null;
                  end case;
                when "00110" =>
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"8A" |		-- TXA
                 x"98" |		-- TYA
                 x"9A" |		-- TXS
                 x"A8" |		-- TAY
                 x"AA" |		-- TAX
                 x"BA" =>		-- TSX
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  Operation <= opTxx;

                  case Opcode is
                    when x"8A" =>
                      Accu(7 downto 0) <= X(7 downto 0);
                      ALU_a <= X(7 downto 0);
                    when x"98" =>
                      Accu(7 downto 0) <= Y(7 downto 0);
                      ALU_a <= Y(7 downto 0);
                    when x"9A" =>
                      SP(7 downto 0) <= X(7 downto 0);
                      ALU_a <= X(7 downto 0);
                    when x"A8" =>
                      Y(7 downto 0) <= Accu(7 downto 0);
                      ALU_a <= Accu(7 downto 0);
                    when x"AA" =>
                      X(7 downto 0) <= Accu(7 downto 0);
                      ALU_a <= Accu(7 downto 0);
                    when x"BA" =>
                      X(7 downto 0) <= SP(7 downto 0);
                      ALU_a <= SP(7 downto 0);
                    when others => null;
                  end case;
                when "00100" =>
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"8C" |		-- STY $xxxx
                 x"8D" |		-- STA $xxxx
                 x"8E" | 		-- STx $xxxx

                 x"99" |		-- STA $xxxx,Y
                 x"9D" =>		-- STA $xxxx,X
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;
                when "00100" =>		-- read LB indexed address
                  PC <= PC + 1;
                  PCl <= IntData;
                when "00110" =>		-- read HB address
                  AddressMode <= amTmp;
                  RW_i <= '0';

                  if (Opcode(4) = '0') then
                    Tmp <= IntData & PCl;
                  elsif (Opcode(2) = '0') then
                    Tmp <= IntData & PCl + Y;
                  else
                    Tmp <= IntData & PCl + X;
                  end if;

                  case Opcode(1 downto 0) is
                    when "00" => D_out <= Y(7 downto 0);
                    when "01" => D_out <= Accu(7 downto 0);
                    when "10" => D_out <= X(7 downto 0);
                    when others => null;
                  end case;
                when "01000" =>
                  AddressMode <= amPC;
                  PC <= PC + 1;
                  RW_i <= '1';

                  if (Opcode(4) = '0') then
                    OpcC_reset <= '1';
                  end if;  
                when "01010" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"91" =>		-- STA ($xx),Y
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;
                when "00100" =>		-- read zero page address
                  AddressMode <= amTmp;
                  Tmp <= x"00" & IntData;
                when "00110" =>		-- read HB indexed address
                  Tmp <= (Tmp + 1) and x"00FF";
                  PCl <= IntData;		-- store LB indexed address
                when "01000" =>		-- read byte at indexed address
                  Tmp <= (IntData  & PCl) + Y;
                  D_out <= Accu(7 downto 0);
                  RW_i <= '0';
                when "01010" =>		-- calculate and store results
                  AddressMode <= amPC;
                  PC <= PC + 1;
                  RW_i <= '1';
                when "01100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"A0" |		-- LDY #nn
                 x"A2" |		-- LDX #nn
                 x"A9" =>		-- LDA #nn
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  Operation <= opLDx;
                when "00100" =>
                  case Opcode(2 downto 0) is
                    when "000" => Y(7 downto 0) <= IntData;
                    when "001" => Accu(7 downto 0) <= IntData;
                    when "010" => X(7 downto 0) <= IntData;
                    when others => null;
                  end case;

                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;
                  PC <= PC + 1;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"C0" |		-- CPY #n
                 x"C9" |		-- CMP #n
                 x"E0" =>		-- CPX #n
              case OpcCounter is
                when "00010" =>		-- read byte
                  PC <= PC + 1;
                  Operation <= opCMP;

                  case Opcode(5 downto 3) is
                    when "000" => ALU_a <= Y(7 downto 0);
                    when "001" => ALU_a <= Accu(7 downto 0);
                    when "100" => ALU_a <= X(7 downto 0);
                    when others => null;		
                  end case;
                when "00100" =>		-- calculate and store results
                  PC <= PC + 1;
                  P(Flag_C) <= ALU_C;
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"C4" |		-- CPY $xx
                 x"C5" |		-- CMP $xx
                 x"D5" |		-- CMP $xx,X
                 x"E4" =>		-- CPX $xx
              case OpcCounter is
                when "00010" =>		-- read byte
                  PC <= PC + 1;
                  Operation <= opCMP;

                  case Opcode(5 downto 0) is
                    when "000100" => ALU_a <= Y(7 downto 0);
                    when "100100" => ALU_a <= X(7 downto 0);
                    when others => ALU_a <= Accu(7 downto 0);
                  end case;
                when "00100" =>		-- calculate and store results
                  AddressMode <= amTmp;

                  if (Opcode = x"D5") then
                    Tmp <= x"00" & (IntData + X(7 downto 0));
                  else
                    Tmp <= x"00" & IntData;
                  end if;
                when "00110" =>		-- calculate and store results
                  AddressMode <= amPC;
                  PC <= PC + 1;
                  P(Flag_C) <= ALU_C;
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"CC" |		-- CPY $xxxx
                 x"CD" |		-- CMP $xxxx
                 x"EC" |		-- CPX $xxxx

                 x"D9" |		-- CMP $xxxx,Y
                 x"DD" =>		-- CMP $xxxx,X
              case OpcCounter is
                when "00010" =>		-- read byte
                  PC <= PC + 1;
                  Operation <= opCMP;

                  case Opcode is
                    when x"CC" => ALU_a <= Y(7 downto 0);
                    when x"EC" => ALU_a <= X(7 downto 0);
                    when others => ALU_a <= Accu(7 downto 0);
                  end case;
                when "00100" =>		-- read LB address
                  PC <= PC + 1;
                  PCl <= IntData;
                when "00110" =>		-- read HB address
                  AddressMode <= amTmp;

                  case Opcode(4 downto 2) is
		    when "111"  => Tmp <= (IntData & PCl) + X;
                    when "110"  => Tmp <= (IntData & PCl) + Y;
                    when others => Tmp <= IntData & PCl;
                  end case;
                when "01000" =>		-- calculate and store results
                  AddressMode <= amPC;
                  PC <= PC + 1;
                  P(Flag_C) <= ALU_C;
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"EA" =>		-- NOP
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                when "00100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;
            when others => null;
          end case;
        end if;
      end if;
    end if;
  end process;



--  ====  ALU  =====

-- Select use of decimal flag
  with Operation select ALU_D <=
    '0'		when opBRA | opCMP,
    P(Flag_D)	when others;

-- Select use of carry flag
  with Operation select ALU_Ci <=
    '0'		when opBRA,
    '1'		when opCMP,
    P(Flag_C)	when others;


-- Set second operand
  with Operation select ALU_b <=
    not IntData		when opSBC | opCMP,
    ALU_a		when opASLa | opLSRa | opROLa | opRORa,
    PC(7 downto 0)	when opBRA,
    IntData		when others;


-- Add 1st and 2nd operand
  NibbleL <=('0' & ALU_a(3 downto 0)) + ('0' & ALU_b(3 downto 0))
              + ("0000" & ALU_Ci);
		
--  Gideon    if (   (NibbleL(4) = '1') or (NibbleL(3 downto 2) = "11")
--	        or (NibbleL(3 downto 1) = "101")) then
  NibbleL2 <= NibbleL + 6 when ((ALU_D = '1') and (NibbleL > 9)) else NibbleL;
     	
  NibbleH <= ('0' & ALU_a(7 downto 4)) + ('0' & ALU_b(7 downto 4))
               + ("0000" & NibbleL2(4));

--  Gideon    if (   (NibbleH(4) = '1') or (NibbleH(3 downto 2) = "11")
--	        or (NibbleH(3 downto 1) = "101")
--		or ((NibbleH(4) = '1') and (NibbleH(3 downto 0) = "1001"))) then -- >9 (10-11, 12-15)
  NibbleH2 <= NibbleH + 6 when ((ALU_D = '1') and (NibbleH > 9)) else NibbleH;


-- Calculate result
  with Operation select ALU_r <=
    NibbleH2(3 downto 0) & NibbleL2(3 downto 0)	when opADC | opSBC | opCMP | opBRA,
    ALU_a and IntData				when opAND | opBIT,
    ALU_b(6 downto 0) & "0"			when opASL | opASLa,
    IntData - 1					when opDEC,
    ALU_a xor IntData				when opEOR,
    IntData + 1					when opINC,
    "0" & ALU_b(7 downto 1)			when opLSR | opLSRa,
    ALU_a or IntData				when opOR,
    ALU_b(6 downto 0) & ALU_Ci			when opROL | opROLa,
    ALU_Ci & ALU_b(7 downto 1)			when opROR | opRORa,
    ALU_a					when opTxx,
    IntData					when others;


-- Calculate Carry flag
  with Operation select ALU_C <=
    ALU_b(7)	when opASL | opROL | opASLa | opROLa,
    ALU_b(0)	when opLSR | opROR | opLSRa | opRORa,
    NibbleH2(4)	when others;	


-- Calculate Negative flag
  with Operation select ALU_N <=
    IntData(7)	when opBIT,
    ALU_r(7)	when others;


-- Calculate Overflow flag
  with Operation select ALU_V <=
    ((ALU_a(7) xor ALU_r(7)) and (ALU_b(7) xor ALU_r(7)))
           	when opADC | opSBC | opCMP | opBRA,
    IntData(6)	when others;


-- Calculate Zero flag
  ALU_Z <= '1'	when ALU_r = 0 else '0';


-- Select whether Branch should be taken or not
  with Opcode(7 downto 5) select Branch <=
    not P(Flag_N)	when "000",		-- BPL
    P(Flag_N) 		when "001",		-- BMI
    not P(Flag_V) 	when "010",		-- BVC
    P(Flag_V) 		when "011",		-- BVS
    not P(Flag_C) 	when "100",		-- BCC
    P(Flag_C) 		when "101",		-- BCS
    not P(Flag_Z) 	when "110",		-- BNE
    P(Flag_Z)	 	when others;		-- DEQ




--IntD <= std_logic_vector(IntData);				--#$#1
--p_i <= PHI_i;							--#$#1
--XXX1 <= std_logic_vector(ALU_a);				--#$#1
--XXX2 <= std_logic_vector(ALU_b);				--#$#1
--XXX3 <= std_logic_vector(ALU_r);				--#$#1


end AR_RB65;

