-- 65816 compatible microprocessor core
--
-- Version : 0.1
--
--

library ieee;
  use ieee.std_logic_1164.all;
  use ieee.numeric_std.all ;


entity RB65816 is
  Port (
    A		: out	std_logic_vector(23 downto 0);
    ABORT	: in	std_logic;
    Din		: in	std_logic_vector(7 downto 0);
    Dout	: out 	std_logic_vector(7 downto 0);
    E		: out	std_logic;    
    IRQ		: in	std_logic;
    MLB		: out	std_logic;			-- Memory Lock    
    MX		: out	std_logic;			-- Memor/indeX select status   
    NMI		: in	std_logic;
    PHI0	: in	std_logic;			-- input single clock
    RDY		: in	std_logic;
    RDYout	: out	std_logic;    
    RESET	: in	std_logic;
    RW		: out	std_logic;			-- Read/Write line
    VDA		: out	std_logic;			-- Valid Data Address
    VPA		: out	std_logic;			-- Valid Program Address
    VPB		: out	std_logic;			-- Vector Pull Bit    

-- Debug purposes
    A2b		: out	std_logic_vector(15 downto 0);
    ALU_a2	: out 	std_logic_vector(15 downto 0);
    ALU_b2	: out 	std_logic_vector(15 downto 0);
    ALU_r2	: out 	std_logic_vector(15 downto 0);
    ALU_C2	: out	std_logic;    
    ALU_V2	: out	std_logic;    
    ALU_Z2	: out	std_logic;    
    AM		: out	std_logic_vector(1 downto 0);
    Intdata2	: out 	std_logic_vector(7 downto 0);
    OpcC	: out	std_logic_vector(4 downto 0);	-- Opcode step counter
xx1,
    Version	: out 	std_logic_vector(7 downto 0)
  );
end RB65816;


architecture AR_RB65816 of RB65816 is
  signal Accu	: unsigned(15 downto 0) := x"FFFF";	-- 16 bits reg. A
  signal DBR	: unsigned(7 downto 0) := x"00";	-- Data Bank Register
  signal DR	: unsigned(15 downto 0) := x"0000";	-- Direct Register
  signal PBR	: unsigned(7 downto 0) := x"00";	-- Program Bank Register
  signal PC	: unsigned(15 downto 0); 		-- Program Counter
  signal SP	: unsigned(15 downto 0) := x"01FF";	-- 16 bits reg. Stack Pointer
  signal X	: unsigned(15 downto 0) := x"0000";	-- 16 bits reg. X
  signal Y	: unsigned(15 downto 0) := x"0000";	-- 16 bits reg. Y

  signal P	: unsigned(7 downto 0) := x"FF";	-- 8 bits Emulation Status reg.
  signal Pn	: unsigned(7 downto 0) := x"FF";	-- 8 bits Native Status reg.

  signal Tmp	: unsigned(15 downto 0);		-- Temporary buffer
  signal Tmp2	: unsigned(15 downto 0);		-- Temporary buffer
  signal TmpAlu	: unsigned(7 downto 0);			-- Temporary buffer


--  ALU  --
  signal ALU_a,						-- input 1 (Accu)
         ALU_b,
         ALU_r		: unsigned(15 downto 0) := x"0000";	-- result

  signal ALU_C,						-- output ALU Carry
         ALU_Ci,					-- input ALU Carry
         ALU_D,						-- input ALU Decimal
         ALU_V,						-- output ALU Overflow
         ALU_Z,						-- output ALU Zero
         ALU_N		: std_logic := '0';		-- output ALU Negative

  signal NibbleL,
         NibbleL2,
         NibbleWL,
         NibbleWL2,
         NibbleH,
         NibbleH2,
         NibbleWH,
	 NibbleWH2	: unsigned(4 downto 0) := "00000";	

  signal Operation	: unsigned(5 downto 0);	-- ALU operation


  signal AddressMsb	: unsigned(1 downto 0) := "10";
  signal AddressMode	: unsigned(1 downto 0) := "10";
  signal D_out		: unsigned(7 downto 0) := x"FF";
  signal IntData    	: unsigned(7 downto 0);		-- Internal data
  signal Opcode    	: unsigned(7 downto 0);		-- Opcode

  signal Branch 	: std_logic := '0';
  signal DoubleAction	: std_logic := '0';		-- use a step twice in case od 16 bits action
  signal IRQ_in		: std_logic := '0';
  signal M16_A		: std_logic := '0';		-- flag for 16-bit mode
  signal M16_X		: std_logic := '0';		-- flag for 16-bit mode
  signal NMI_busy	: std_logic := '0';
  signal NMI_handled	: std_logic := '1';
  signal NMI_in		: std_logic := '0';
  signal OpcC_reset	: std_logic := '0';
  signal Reset_in	: std_logic := '0';		-- RESET detected
  signal RW_i		: std_logic;			-- 0 = write, 1 = read
  signal WAI_flg	: std_logic := '0';


  constant ahZero	: unsigned(1 downto 0) := "00";
  constant ahPBR	: unsigned(1 downto 0) := "01";
  constant ahDBR	: unsigned(1 downto 0) := "10";

  constant amSP		: unsigned(1 downto 0) := "00";
  constant amPC		: unsigned(1 downto 0) := "01";
  constant amTmp	: unsigned(1 downto 0) := "10";
  constant amTmp2	: unsigned(1 downto 0) := "11";


  constant Flag_C 	: integer := 0;
  constant Flag_Z 	: integer := 1;
  constant Flag_I 	: integer := 2;
  constant Flag_D 	: integer := 3;
  constant Flag_B 	: integer := 4;
  constant Flag_1 	: integer := 5;
  constant Flag_V 	: integer := 6;
  constant Flag_N 	: integer := 7;

  constant Flag_Emu 	: integer := 0;
  constant Flag_X16	: integer := 4;
  constant Flag_A16	: integer := 5;
  

  constant opAND 	: unsigned(5 downto 0) := "000000";
  constant opOR  	: unsigned(5 downto 0) := "000001";
  constant opEOR 	: unsigned(5 downto 0) := "000010";
  constant opASL 	: unsigned(5 downto 0) := "000011";
  constant opLSR 	: unsigned(5 downto 0) := "000100";
  constant opROL 	: unsigned(5 downto 0) := "000101";
  constant opROR 	: unsigned(5 downto 0) := "000110";
  constant opINC 	: unsigned(5 downto 0) := "000111";
  constant opDEC 	: unsigned(5 downto 0) := "001000";
  constant opADC 	: unsigned(5 downto 0) := "001001";
  constant opCMP 	: unsigned(5 downto 0) := "001010";
  constant opBIT 	: unsigned(5 downto 0) := "001011";
  constant opSBC	: unsigned(5 downto 0) := "001100";
  constant opBRA 	: unsigned(5 downto 0) := "001101";
  constant opLDx 	: unsigned(5 downto 0) := "001110";
  constant opTxx 	: unsigned(5 downto 0) := "001111";
  constant opTRB 	: unsigned(5 downto 0) := "010000";
  constant opTSB 	: unsigned(5 downto 0) := "010001";
  constant opASLa 	: unsigned(5 downto 0) := "010011";
  constant opLSRa 	: unsigned(5 downto 0) := "010100";
  constant opROLa 	: unsigned(5 downto 0) := "010101";
  constant opRORa 	: unsigned(5 downto 0) := "010110";

  constant opwAND 	: unsigned(5 downto 0) := "100000";
  constant opwOR  	: unsigned(5 downto 0) := "100001";
  constant opwEOR 	: unsigned(5 downto 0) := "100010";
  constant opwASL 	: unsigned(5 downto 0) := "100011";
  constant opwLSR 	: unsigned(5 downto 0) := "100100";
  constant opwROL 	: unsigned(5 downto 0) := "100101";
  constant opwROR 	: unsigned(5 downto 0) := "100110";
  constant opwINC 	: unsigned(5 downto 0) := "100111";
  constant opwDEC 	: unsigned(5 downto 0) := "101000";
  constant opwADC 	: unsigned(5 downto 0) := "101001";
  constant opwCMP 	: unsigned(5 downto 0) := "101010";
  constant opwBIT 	: unsigned(5 downto 0) := "101011";
  constant opwSBC	: unsigned(5 downto 0) := "101100";
  constant opwBRA 	: unsigned(5 downto 0) := "101101";
  constant opwLDx 	: unsigned(5 downto 0) := "101110";
  constant opwTxx 	: unsigned(5 downto 0) := "101111";
  constant opwTRB 	: unsigned(5 downto 0) := "110000";
  constant opwTSB 	: unsigned(5 downto 0) := "110001";
  constant opwASLa 	: unsigned(5 downto 0) := "110011";
  constant opwLSRa 	: unsigned(5 downto 0) := "110100";
  constant opwROLa 	: unsigned(5 downto 0) := "110101";
  constant opwRORa 	: unsigned(5 downto 0) := "110110";

  constant opNULL	: unsigned(5 downto 0) := "111111";

			
begin
  Version <= x"0C";

  Dout <= std_logic_vector(D_out);
  E <= Pn(Flag_Emu);
  IntData <= unsigned(Din) when (PHI0 = '0');				-- read data
  M16_A <= Pn(Flag_Emu) or Pn(Flag_A16);
  M16_X <= Pn(Flag_Emu) or Pn(Flag_X16);
  Opcode <= unsigned(Din) when ((PHI0 = '0') and (OpcC_Reset = '1'));
  RW <= RW_i;


  with PHI0 select MX <=
    Pn(Flag_A16)			when '0',
    Pn(Flag_X16)			when others;


  with AddressMsb select A(23 downto 16) <=
    std_logic_vector(DBR)		when ahDBR,
    std_logic_vector(PBR)		when ahPBR,
    x"00"				when others;


  with AddressMode select A(15 downto 0) <=
    std_logic_vector(PC)		when amPC,
    std_logic_vector(SP)		when amSP,
    std_logic_vector(Tmp)		when amTmp,
    std_logic_vector(Tmp2)		when others;

A2b <= std_logic_vector(PC);
ALU_a2 <= std_logic_vector(ALU_a);
ALU_b2 <= std_logic_vector(ALU_b);
ALU_r2 <= std_logic_vector(ALU_r);
ALU_C2 <= std_logic(ALU_C);
ALU_V2 <= std_logic(ALU_V);
ALU_Z2 <= std_logic(ALU_Z);
AM <= std_logic_vector(AddressMode);
Intdata2 <= std_logic_vector(Intdata);


  process(NMI)
  begin
    if ((NMI = '1') and (NMI_in = '1') and (NMI_handled = '1')) then
      NMI_in <= '0';
    elsif (falling_edge(NMI) and (NMI_in = '0')) then
      NMI_in <= '1';
    end if;
  end process;


  process(RESET, PHI0)
    variable OpcCounter : unsigned(4 downto 0) := "00000";	-- Opcode step counter, PHI2 is LSB
  begin
    if (RESET = '0') then
      AddressMode <= amTmp;
      AddressMSB <= ahZero;
      DBR <= x"00";
      DR <= x"0000";
      NMI_busy <= '0';
      NMI_handled <= '1';
      OpcC <= std_logic_vector(OpcCounter);
      OpcC_reset <= '1';
      Operation <= opNULL;
      PBR <= x"00";
      PC <= x"FFFC";
      Pn(Flag_Emu) <= '1';
      Pn(Flag_A16) <= '1';
      Pn(Flag_X16) <= '1';
      Reset_in <= '1';
      RDYout <= '1';
      RW_i <= '1';
      SP <= x"01FF";
      Tmp <= x"FFFC";
      VDA <= '0';
      VPA <= '0';
      VPB <= '1';
      WAI_flg <= '0';
      X <= x"0000";
      Y <= x"0000";
 xx1 <= x"00";
    elsif rising_edge(PHI0) then

--  Check if RDY input is negated
      if (RDY = '1') then

--  Handle possible NMI/IRQ at start of every instruction
        if (OpcC_Reset = '1') then
          OpcCounter := "00010";
          OpcC_Reset <= '0';

          if ((NMI_in = '1') or (IRQ = '0')) then
            WAI_flg <= '0';
          end if;
	      
          if (   (NMI_in = '1')
              or ((IRQ = '0') and (P(Flag_I) = '0'))
             ) then			-- push HB PC to Stack
            AddressMode <= amSP;
            AddressMSB <= ahZero;
            RW_i <= '0';
            D_out <= PC(15 downto 8);

            if (NMI_in = '1') then
              NMI_busy <= '1';
              NMI_handled <= '0';
            else
             IRQ_in <= '1';
            end if;
          else
            AddressMode <= amPC;
            AddressMSB <= ahPBR;
          end if;
        else
          OpcCounter := OpcCounter + 2;
        end if;

        OpcC <= std_logic_vector(OpcCounter);

--  === Reset ===
        if (Reset_in = '1') then
          case OpcCounter is
            when "00010" =>
              AddressMode <= amTmp;
              Tmp <= x"FFFC";
              IRQ_in <= '0';
              RW_i <= '1';
            when "00100" =>
              PC(7 downto 0) <= IntData;	-- read low byte reset routine
              Tmp <= x"FFFD";
            when "00110" =>
              AddressMode <= amPC; 	-- output read vector
              AddressMSB <= ahPBR;
              PC(15 downto 8) <= IntData;	-- read high byte reset routine
              NMI_busy <= '0';
              NMI_handled <= '1';
              OpcC_reset <= '1';
              Reset_in <= '0';
            when others => null;
          end case;


--  === IRQ / NMI ===
        elsif ((NMI_busy = '1') or (IRQ_in = '1')) then
          case OpcCounter is
            when "00100" =>			-- push LB PC to Stack
              SP <= SP - 1;
              
              if (NMI_busy = '0') then
                P(Flag_B) <= '0';
              end if;
              
	      D_out <= PC(7 downto 0);
              RW_i <= '0';
            when "00110" =>			-- push Status Register to Stack
              SP <= SP - 1;
              D_out <= P;
              RW_i <= '0';
            when "01000" =>
              AddressMode <= amTmp;
              if (NMI_busy = '1') then
                Tmp <= x"FFFA";
              else
                Tmp <= x"FFFE";
              end if;
              P(Flag_I) <= '0';
              RW_i <= '1';
            when "01010" =>
              PC(7 downto 0) <= IntData;	-- read low byte routine
              if (NMI_busy = '1') then
                Tmp <= x"FFFB";
              else
                Tmp <= x"FFFF";
              end if;
            when "01100" =>
              AddressMode <= amPC;	 	-- output read vector
              AddressMSB <= ahPBR;
              PC(15 downto 8) <= IntData;	-- read high byte routine
            when "01110" =>		-- COMPATIBILITY
              if (IRQ_in = '1') then
                IRQ_in <= '0';
              else
                NMI_busy <= '0';
                NMI_handled <= '1';
              end if;
              OpcC_reset <= '1';
            when others => null;
          end case;


--  === All opcodes ===
        else
          case Opcode is

            when x"00" |		-- BRK
                 x"02" =>		-- COP          
              case OpcCounter is
                when "00010" =>			-- increment PC twice
                  AddressMode <= amSP;
                  AddressMSB <= ahZero;
                  PC <= PC + 2;
                  RW_i <= '0';
                  D_out <= PC(15 downto 8);	-- push HB PC to Stack
                when "00100" =>			-- push LB PC to Stack
                  SP <= SP - 1;
                  RW_i <= '0';
                  D_out <= PC(7 downto 0);
                when "00110" =>			-- push Status Register to Stack
                  SP <= SP - 1;
                  RW_i <= '0';
                  D_out <= P;
                when "01000" =>
                  AddressMode <= amTmp;
                  RW_i <= '1';

                  if (Opcode(1) = '0') then
                    if (Pn(Flag_Emu) = '1') then
                      Tmp <= x"FFFE";		-- emulation mode vector
                      P(Flag_B) <= '1';
                    else  
                      Tmp <= x"FFF8";		-- native mode vector
                    end if; 
                  else
                    Tmp <= x"FFF4";
		  end if;  
                when "01010" =>
                  PC(7 downto 0) <= IntData;	-- read low byte BRK routine
                  Tmp <= Tmp + 1;
                when "01100" =>
                  AddressMode <= amPC;		-- store read vector into PC
                  AddressMSB <= ahPBR;
                  PC(15 downto 8) <= IntData;	-- read high byte BRK routine
                when "01110" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"01" |		-- ORA ($xx,X)
                 x"21" |		-- AND ($xx,X)
                 x"41" |		-- EOR ($xx,X)
                 x"61" |		-- ADC ($xx,X)
                 x"81" |		-- STA ($xx,X)
                 x"A1" |		-- LDA ($xx,X)
                 x"C1" |		-- CMP ($xx,X)
                 x"E1" =>		-- SBC ($xx,X)
              case OpcCounter is
                when "00010" =>			-- read zero page address
                  PC <= PC + 1;

                  case Opcode(7 downto 5) is
                    when "000" => Operation <= opOR;
                    when "001" => Operation <= opAND;
                    when "010" => Operation <= opEOR;
                    when "011" => Operation <= opADC;
                    when "101" => Operation <= opLDx;
                    when "110" => Operation <= opCMP;
                    when "111" => Operation <= opSBC;
                    when others => null;
                  end case;
                when "00100" =>			-- read LB indexed address
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;
                  PC <= PC + 1;
                  
                  if (M16_X = '0') then
                    Tmp <= DR + IntData + X;
                  else
                    Tmp <= DR + (x"00" & (IntData + X(7 downto 0)));
                  end if;
                when "00110" =>			-- read HB indexed address
                  Tmp <= Tmp + 1;
                  Tmp2(7 downto 0) <= IntData;	-- store LB indexed address
                when "01000" =>			-- read byte at indexed address
                  ALU_a <= Accu;
                  AddressMode <= amTmp2;
                  Tmp2(15 downto 8) <= IntData;	-- store HB indexed address
                  
                  if (Opcode(7 downto 5) = "100") then
                    D_out <= Accu(7 downto 0);
                    RW_i <= '0';
                  end if;

                  if (M16_A = '0') then
                    DoubleAction <= '1';
                    Operation <= '1' & Operation(4 downto 0);
                  end if;
                when "01010" =>		-- calculate and store results
                  if (DoubleAction = '1') then
                    if (Opcode(7 downto 5) = "100") then
                      D_out <= Accu(15 downto 8);
                    else  
                      TmpAlu <= Intdata;
                    end if;

                    OpcCounter := OpcCounter - 2;	-- "01000"
                    DoubleAction <= '0';
                    Tmp2 <= Tmp2 + 1;
                  else  
                    AddressMode <= amPC;
                    AddressMSB <= ahPBR;

                    if (    (Opcode(7 downto 5) /= "100")
		        and (Opcode(7 downto 5) /= "110")) then
                      if (M16_A = '0') then
                        Accu <= ALU_r;
                      else  
                        Accu(7 downto 0) <= ALU_r(7 downto 0);
                      end if;  
                    end if;
                  end if; 

                  if (Opcode(7 downto 5) = "100") then
                    RW_i <= '1';
                  else
                    P(Flag_N) <= ALU_N;
                    P(Flag_Z) <= ALU_Z;

                    case Opcode(6 downto 5) is
                      when "11" => P(Flag_V) <= ALU_V;
                      when others => null;
                    end case;

                    case Opcode(7 downto 5) is
                      when "011" | "110" | "111" => P(Flag_C) <= ALU_C;
                      when others => null;
                    end case;
                  end if;
                when "01100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"04" |		-- TSB $xx
                 x"14" =>		-- TRB $xx
              case OpcCounter is
                when "00010" =>			-- read zero page address
                  Operation <= opBIT;
                  PC <= PC + 1;
                when "00100" =>			-- read zero page address
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;
                  Tmp <= DR + IntData;

                  if (M16_A = '0') then
                    DoubleAction <= '1';
                    Operation <= '1' & Operation(4 downto 0);
                  end if;
                when "00110" =>	
                  if (DoubleAction = '1') then
                    TmpAlu <= Intdata;
                    OpcCounter := OpcCounter - 2;	-- "00100"
                    DoubleAction <= '0';
                  else  
                    if (M16_A = '0') then
                      DoubleAction <= '1';
                    end if;
                  end if;

                  P(Flag_Z) <= ALU_Z;
                when "01000" =>
                  RW_i <= '0';

                  if (Opcode(4) = '0') then
                    Operation <= opTSB;
                  else  
                    Operation <= opTRB;
                  end if;
                  
                  if (DoubleAction = '1') then
                    D_out <= ALU_r(7 downto 0);		--  output altered value
                    OpcCounter := OpcCounter - 2;	-- "00110"
                    DoubleAction <= '0';
                  else  
                    if (M16_A = '0') then
                      Operation <= '1' & Operation(4 downto 0);
                      D_out <= ALU_r(15 downto 8);	--  output altered value
                    else  
                      D_out <= ALU_r(7 downto 0);	--  output altered value
                    end if;
                  end if;
                when "01010" =>
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  PC <= PC + 1;
                  RW_i <= '1';

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"05" |		-- ORA $xx
                 x"25" |		-- AND $xx
                 x"45" |		-- EOR $xx
                 x"65" |		-- ADC $xx
                 x"A5" |		-- LDA $xx
                 x"E5" |		-- SBC $xx

                 x"15" |		-- ORA $xx,X
                 x"35" |		-- AND $xx,X
                 x"55" |		-- EOR $xx,X
                 x"75" |		-- ADC $xx,X
                 x"B5" |		-- LDA $xx,X
                 x"F5" =>		-- SBC $xx,X
              case OpcCounter is
                when "00010" =>			-- read zero page address
                  PC <= PC + 1;
                  ALU_a <= Accu;

                  case Opcode(7 downto 5) is
                    when "000" => Operation <= opOR;
                    when "001" => Operation <= opAND;
                    when "010" => Operation <= opEOR;
                    when "011" => Operation <= opADC;
                    when "101" => Operation <= opLDx;
                    when "111" => Operation <= opSBC;
                    when others => null;
                  end case;
                when "00100" =>			-- read zero page address
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;
                  PC <= PC + 1;

                  if (Opcode(4) = '0') then		-- $xx
                    Tmp <= DR + IntData;
                  elsif (Opcode(1) = '0') then		-- $xx,X
                    if (M16_X = '0') then
                      Tmp <= DR + IntData + X;
                    else
                      Tmp <= DR + (x"00" & (IntData + X(7 downto 0)));
                    end if;
                  end if;

                  if (M16_A = '0') then
                    DoubleAction <= '1';
                    Operation <= '1' & Operation(4 downto 0);
                  end if;
                when "00110" =>		-- calculate and store results
                  if (DoubleAction = '1') then
                    TmpAlu <= Intdata;
                    OpcCounter := OpcCounter - 2;	-- "01000"
                    DoubleAction <= '0';
                    Tmp <= Tmp + 1;
                  else  
                    AddressMode <= amPC;
                    AddressMSB <= ahPBR;

                    if (M16_A = '0') then
                      Accu <= ALU_r;
                    else 
                      Accu(7 downto 0) <= ALU_r(7 downto 0);
                    end if;
                  end if; 

                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;

                  if (    (Opcode(6 downto 5) = "11")
		      and (Opcode(3 downto 0) = "0101")) then
                    P(Flag_V) <= ALU_V;
                    P(Flag_C) <= ALU_C;
                  end if;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"A4" |		-- LDY $xx
                 x"A6" |		-- LDX $xx
                 x"B4" |		-- LDY $xx,X
                 x"B6" =>		-- LDX $xx,Y
              case OpcCounter is
                when "00010" =>			-- read zero page address
                  PC <= PC + 1;
                  Operation <= opLDx;
                when "00100" =>			-- read zero page address
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;
                  PC <= PC + 1;

                  if (Opcode(4) = '0') then		-- $xx
                    Tmp <= DR + IntData;
                  elsif (Opcode(1) = '0') then		-- $xx,X
                    if (M16_X = '0') then
                      Tmp <= DR + IntData + X;
                    else
                      Tmp <= DR + (x"00" & (IntData + X(7 downto 0)));
                    end if;
                  else					-- $xx,Y
                    if (M16_X = '0') then
                      Tmp <= DR + IntData + Y;
                    else
                      Tmp <= DR + (x"00" & (IntData + Y(7 downto 0)));
                    end if;
                  end if;

                  if (M16_X = '0') then
                    DoubleAction <= '1';
                    Operation <= '1' & Operation(4 downto 0);
                  end if;
                when "00110" =>		-- calculate and store results
                  if (DoubleAction = '1') then
                    TmpAlu <= Intdata;
                    OpcCounter := OpcCounter - 2;	-- "00100"
                    DoubleAction <= '0';
                    Tmp <= Tmp + 1;
                  else  
                    AddressMode <= amPC;
                    AddressMSB <= ahPBR;

                    if (Opcode(1) = '0') then
                      Y <= ALU_r;
                    else  
                      X <= ALU_r;
-- remark: high byte can be ignored. Will be set to zero anyway when set to
--         native mode.

                    end if;
                  end if; 

                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"06" |		-- ASL $xx
                 x"26" |		-- ROL $xx
                 x"46" |		-- LSR $xx
                 x"66" |		-- ROR $xx
                 x"C6" |		-- DEC $xx
                 x"E6" | 		-- INC $xx

                 x"16" |		-- ASL $xx,X
                 x"36" |		-- ROL $xx,X
                 x"56" |		-- LSR $xx,X
                 x"76" |		-- ROR $xx,X
                 x"D6" |		-- DEC $xx,X
                 x"F6" =>		-- INC $xx,X
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  case Opcode(7 downto 5) is
                    when "000" => Operation <= opASL;
                    when "001" => Operation <= opROL;
                    when "010" => Operation <= opLSR;
                    when "011" => Operation <= opROR;
                    when "110" => Operation <= opDEC;
                    when "111" => Operation <= opINC;
                    when others => null;
                  end case;

                  PC <= PC + 1;
                when "00100" =>		-- read LB indexed address
                  PC <= PC + 1;
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;

                  if (Opcode(4) = '0') then
                    Tmp <= DR + IntData;
                  else 
                    if (M16_X = '0') then
                      Tmp <= DR + IntData + X;
                    else
                      Tmp <= DR + (x"00" & (IntData + X(7 downto 0)));
                    end if;
                  end if;

                  if (M16_A = '0') then
                    DoubleAction <= '1';
                    Operation <= '1' & Operation(4 downto 0);
                  end if;
                when "00110" =>		-- calculate and store results
                  if (DoubleAction = '1') then
                    TmpAlu <= Intdata;
                    OpcCounter := OpcCounter - 2;	-- "00100"
                    DoubleAction <= '0';
                    Tmp <= Tmp + 1;
                  else  
                    RW_i <= '0';
                    P(Flag_N) <= ALU_N;
                    P(Flag_Z) <= ALU_Z;
                    D_out <= ALU_r(7 downto 0);

                    if (Opcode(7) = '0') then
                      P(Flag_C) <= ALU_C;
                    end if;

                    if (M16_A = '0') then
                      DoubleAction <= '1';
                    end if;
                  end if;  
                when "01000" =>
                  if (DoubleAction = '1') then
                    D_out <= ALU_r(15 downto 8);
                    OpcCounter := OpcCounter - 2;	-- "00110"
                    DoubleAction <= '0';
                  else  
                    AddressMode <= amPC;
                    AddressMSB <= ahPBR;
                    RW_i <= '1';
                  end if;  
                when "01010" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"08" |		-- PHP
                 x"0B" |		-- PHD
                 x"48" |		-- PHA
                 x"4B" |		-- PHK
                 x"5A" |		-- PHY
                 x"8B" |		-- PHK = Push DBR
                 x"DA" =>		-- PHX
              case OpcCounter is
                when "00010" =>		-- increment PC / push to stack
                  AddressMode <= amSP;
                  AddressMSB <= ahZero;
                  PC <= PC + 1;
                  RW_i <= '0';

                  Case Opcode is
                    when x"08" => D_out <= P or x"10";
                    when x"0B" => D_out <= DR(7 downto 0);
                    when x"48" => D_out <= Accu(7 downto 0);
                    when x"4B" => D_out <= PBR;
                    when x"5A" => D_out <= Y(7 downto 0);
                    when x"8B" => D_out <= DBR;
                    when x"DA" => D_out <= X(7 downto 0);
                    when others => null;
                  end case;

                  if (
                         (M16_A = '0') 
                      or (M16_X = '0')
		      or (Opcode = x"0B")) then
                    DoubleAction <= '1';
                  end if;
                when "00100" =>
                  if (DoubleAction = '1') then
                    if (Opcode = x"48") then
                      D_out <= DR(15 downto 8);
                    elsif (M16_A = '0') then
                      D_out <= Accu(15 downto 8);
                    else
                      if (Opcode(7) = '0') then
                        D_out <= X(15 downto 8);
                      else
                        D_out <= Y(15 downto 8);
                      end if;

		    end if;

                    SP <= SP - 1;
                    OpcCounter := OpcCounter - 2;	-- "00010"
                    DoubleAction <= '0';
		  else    
                    AddressMode <= amPC;
                    AddressMSB <= ahPBR;
                    SP <= SP - 1;
                    RW_i <= '1';
                  end if;  

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"09" |		-- ORA #n
                 x"29" |		-- AND #n
                 x"49" |		-- EOR #n
                 x"69" |		-- ADC #n
                 x"E9" =>		-- SBC #n
              case OpcCounter is
                when "00010" =>		-- read byte
                  ALU_a <= Accu;
                  PC <= PC + 1;

                  if (M16_A = '0') then
                    case Opcode(7 downto 5) is
                      when "000" => Operation <= opwOR;
                      when "001" => Operation <= opwAND;
                      when "010" => Operation <= opwEOR;
                      when "011" => Operation <= opwADC;
                      when "111" => Operation <= opwSBC;
                      when others => null;
                    end case;

                    DoubleAction <= '1';
                  else  
                    case Opcode(7 downto 5) is
                      when "000" => Operation <= opOR;
                      when "001" => Operation <= opAND;
                      when "010" => Operation <= opEOR;
                      when "011" => Operation <= opADC;
                      when "111" => Operation <= opSBC;
                      when others => null;
                    end case;
                  end if;
                when "00100" =>		-- calculate and store results
                  if (DoubleAction = '1') then
                    TmpAlu <= Intdata;
                    OpcCounter := OpcCounter - 2;	-- "01000"
                    DoubleAction <= '0';
                    PC <= PC + 1;
                  else  
                    if (Opcode(6 downto 5) = "11") then
                      P(Flag_V) <= ALU_V;
                      P(Flag_C) <= ALU_C;
                    end if;

                    if (M16_A = '0') then
                      Accu <= ALU_r;
                    else 
                      Accu(7 downto 0) <= ALU_r(7 downto 0);
                    end if;

                    P(Flag_N) <= ALU_N;
                    P(Flag_Z) <= ALU_Z;
                    PC <= PC + 1;
                  end if;  

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"0A" |		-- ASL A
                 x"2A" |		-- ROL A
                 x"4A" |		-- LSR A
                 x"6A" =>		-- ROR A
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  ALU_a <= Accu;
                  PC <= PC + 1;

                  if (M16_A = '0') then
                    case Opcode(6 downto 5) is
                      when "00" => Operation <= opwASLa;
                      when "01" => Operation <= opwROLa;
                      when "10" => Operation <= opwLSRa;
                      when "11" => Operation <= opwRORa;
                      when others => null;
                    end case;
                  else  
                    case Opcode(6 downto 5) is
                      when "00" => Operation <= opASLa;
                      when "01" => Operation <= opROLa;
                      when "10" => Operation <= opLSRa;
                      when "11" => Operation <= opRORa;
                      when others => null;
                    end case;
                  end if;  
                when "00100" =>		-- calculate and store results
                  if (M16_A = '0') then
                    Accu <= ALU_r;
                  else  
                    Accu(7 downto 0) <= ALU_r(7 downto 0);
                  end if;
                  
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;
                  P(Flag_C) <= ALU_C;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"0C" |		-- TSB $xxxx
                 x"1C" =>		-- TRB $xxxx
              case OpcCounter is
                when "00010" =>		-- read LB address
                  PC <= PC + 1;
                  Operation <= opBIT;
                when "00100" =>		-- read HB address
                  PC <= PC + 1;
                  Tmp(7 downto 0) <= IntData;	-- store LB address
                when "00110" =>		-- store HB address
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;
                  ALU_a <= Accu;
                  Tmp(15 downto 8) <= IntData;

                  if (M16_A = '0') then
                    DoubleAction <= '1';
                    Operation <= '1' & Operation(4 downto 0);
                  end if;
                when "01000" =>		-- calculate Z flag
                  if (DoubleAction = '1') then
                    TmpAlu <= Intdata;
                    OpcCounter := OpcCounter - 2;	-- "00110"
                    DoubleAction <= '0';
                  else  
                    if (M16_A = '0') then
                      DoubleAction <= '1';
                    end if;
                  end if;

                  P(Flag_Z) <= ALU_Z;

-- #### can be ommited?
                when "01010" =>

                  RW_i <= '0';

                  if (Opcode(4) = '0') then
                    Operation <= opTSB;
                  else  
                    Operation <= opTRB;
                  end if;
                  
                  if (DoubleAction = '1') then
                    D_out <= ALU_r(7 downto 0);	--  output altered value
                    OpcCounter := OpcCounter - 2;	-- "00110"
                    DoubleAction <= '0';
                  else  
                    if (M16_A = '0') then
                      Operation <= '1' & Operation(4 downto 0);
                      D_out <= ALU_r(15 downto 8);	--  output altered value
                    else  
                      D_out <= ALU_r(7 downto 0);	--  output altered value
                    end if;
                  end if;
                when "01100" =>
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  PC <= PC + 1;
                  RW_i <= '1';

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"0D" |		-- ORA $xxxx
                 x"2D" |		-- AND $xxxx
                 x"4D" |		-- EOR $xxxx
                 x"6D" |		-- ADC $xxxx
                 x"AC" |		-- LDY $xxxx
                 x"AD" |		-- LDA $xxxx
                 x"AE" |		-- LDX $xxxx
                 x"ED" |		-- SBC $xxxx

                 x"1D" |		-- ORA $xxxx,X
                 x"3D" |		-- AND $xxxx,X
                 x"5D" |		-- EOR $xxxx,X
                 x"7D" |		-- ADC $xxxx,X
                 x"FD" |		-- SBC $xxxx,X
                 x"BC" |		-- LDY $xxxx,X
                 x"BD" |		-- LDA $xxxx,X

                 x"B9" |		-- LDA $xxxx,Y
                 x"BE" |		-- LDX $xxxx,Y
                 x"19" |		-- ORA $xxxx,Y
                 x"39" |		-- AND $xxxx,Y
                 x"59" |		-- EOR $xxxx,Y
                 x"79" |		-- ADC $xxxx,Y
                 x"F9" =>		-- SBC $xxxx,Y
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;

                  case Opcode(7 downto 5) is
                    when "000" => Operation <= opOR;
                    when "001" => Operation <= opAND;
                    when "010" => Operation <= opEOR;
                    when "011" => Operation <= opADC;
                    when "101" => Operation <= opLDx;
                    when "111" => Operation <= opSBC;
                    when others => null;
                  end case;
                when "00100" =>		-- read LB address
                  Tmp2(7 downto 0) <= IntData;
                  PC <= PC + 1;
                  
                when "00110" =>		-- read HB address
                  ALU_a <= Accu;
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;

                  if (Opcode(4) = '0') then
                    Tmp <= IntData & Tmp2(7 downto 0);
		  elsif (Opcode(2 downto 1) = "10") then
                    Tmp <= (IntData & Tmp2(7 downto 0)) + X;
		  else
                    Tmp <= (IntData & Tmp2(7 downto 0)) + Y;
                  end if;

                  if (M16_A = '0') then
                    DoubleAction <= '1';
                    Operation <= '1' & Operation(4 downto 0);
                  end if;
                when "01000" =>		-- calculate and store results
                  if (DoubleAction = '1') then
                    TmpAlu <= Intdata;
                    OpcCounter := OpcCounter - 2;	-- "00110"
                    DoubleAction <= '0';
                    Tmp <= Tmp + 1;
                  else  
                    AddressMode <= amPC;
                    AddressMSB <= ahPBR;
                    PC <= PC + 1;
                    P(Flag_N) <= ALU_N;
                    P(Flag_Z) <= ALU_Z;

                    case Opcode(6 downto 5) is
                      when "11" =>
                        P(Flag_V) <= ALU_V;
                        P(Flag_C) <= ALU_C;
                      when others => null;
                    end case;

                    if (M16_A = '0') then
                      Accu <= ALU_r;
                    else 
                      Accu(7 downto 0) <= ALU_r(7 downto 0);
                    end if;

                    OpcC_reset <= '1';
                  end if;
                when others => null;
              end case;


            when x"0E" |		-- ASL $xxxx
                 x"2E" |		-- ROL $xxxx
                 x"4E" |		-- LSR $xxxx
                 x"6E" |		-- ROR $xxxx
                 x"CE" |		-- DEC $xxxx
                 x"EE" |		-- INC $xxxx

                 x"1E" |		-- ASL $xxxx,X
                 x"3E" |		-- ROL $xxxx,X
                 x"5E" |		-- LSR $xxxx,X
                 x"7E" |		-- ROR $xxxx,X
                 x"DE" |		-- DEC $xxxx,X
                 x"FE" =>		-- INC $xxxx,X
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;
                when "00100" =>		-- read LB address
                  PC <= PC + 1;
                  Tmp2(7 downto 0) <= IntData;
                when "00110" =>		-- read HB address
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;

                  if (Opcode(4) = '0') then
                    Tmp <= IntData & Tmp2(7 downto 0);
		  else
                    Tmp <= (IntData & Tmp2(7 downto 0)) + X;
                  end if;

                  case Opcode(7 downto 5) is
                    when "000" => Operation <= opASL;
                    when "001" => Operation <= opROL;
                    when "010" => Operation <= opLSR;
                    when "011" => Operation <= opROR;
                    when "110" => Operation <= opDEC;
                    when "111" => Operation <= opINC;
                    when others => null;
                  end case;
                when "01000" =>		-- calculate and store results
                  RW_i <= '0';
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;
                  D_out <= ALU_r(7 downto 0);

                  if (Opcode < x"C6") then
                    P(Flag_C) <= ALU_C;
                  end if;
                when "01010" =>
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  PC <= PC + 1;
                  RW_i <= '1';
                when "01100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"10" |		-- BPL
                 x"30" |		-- BMI
                 x"50" |		-- BVC
                 x"70" |		-- BVS
                 x"80" |		-- BRA
                 x"90" |		-- BCC
                 x"B0" |		-- BCS
                 x"D0" |		-- BNE
                 x"F0" =>		-- BEQ
              case OpcCounter is
                when "00010" =>			-- read offset
                  PC <= PC + 1;
                  Operation <= opBRA;
                when "00100" =>			-- decide if branch is needed
                  ALU_a(7 downto 0) <= IntData;	-- store for ALU
                  PC <= PC + 1;			-- increase PC

                  if (Branch = '0') then
                    OpcC_reset <= '1';		-- no branch taken
                  end if;
                when "00110" =>			-- branch taken
                  PC(7 downto 0) <= ALU_r(7 downto 0);

    	-- page boundary crossed
                  if ((ALU_a(7) = '0') and (ALU_C = '1'))then
                    PC(15 downto 8) <= PC(15 downto 8) + 1;
                  elsif ((ALU_a(7) = '1') and (ALU_C = '0'))then
                    PC(15 downto 8) <= PC(15 downto 8) - 1;
                  else  
                    OpcC_reset <= '1';		-- no page boundary crossed
                  end if;
                when "01000" =>			-- page boundary crossd COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"11" |		-- ORA ($xx),Y
                 x"31" |		-- AND ($xx),Y
                 x"51" |		-- EOR ($xx),Y
                 x"71" |		-- ADC ($xx),Y
                 x"B1" |		-- LDA ($xx),Y
                 x"D1" |		-- CMP ($xx),Y
                 x"F1" | 		-- SBC ($xx),Y

                 x"12" |		-- ORA ($xx)
                 x"32" |		-- AND ($xx)
                 x"52" |		-- EOR ($xx)
                 x"72" |		-- ADC ($xx)
                 x"D2" |		-- CMP ($xx)
                 x"F2" =>		-- SBC ($xx)
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;

                  case Opcode(7 downto 5) is
                    when "000" => Operation <= opOR;
                    when "001" => Operation <= opAND;
                    when "010" => Operation <= opEOR;
                    when "011" => Operation <= opADC;
                    when "110" => Operation <= opCMP;
                    when "111" => Operation <= opSBC;
                    when others => null;
                  end case;
                when "00100" =>		-- read zero page address
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;
                  ALU_a <= Accu;
                  Tmp <= DR + IntData;
                when "00110" =>		-- read HB indexed address
                  Tmp2(7 downto 0) <= IntData;		-- store LB indexed address
                  Tmp <= (Tmp + 1) and x"00FF";
                when "01000" =>		-- read byte at indexed address
                  if (Opcode(0) = '1') then
                    Tmp <= (IntData  & Tmp2(7 downto 0)) + Y;
                  else
                    Tmp <= IntData  & Tmp2(7 downto 0);
                  end if;
		    
                  ALU_a <= Accu;
                when "01010" =>		-- calculate and store results
                  AddressMode <= amPC;
                  PC <= PC + 1;
                  AddressMSB <= ahPBR;
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;

                  case Opcode(6 downto 5) is
                    when "11" => P(Flag_V) <= ALU_V;
                    when others => null;
                  end case;

                  case Opcode(7 downto 5) is
                    when "011" | "110" | "111" => P(Flag_C) <= ALU_C;
                    when others => null;
                  end case;

                  if (Opcode /= x"C1") then
                    Accu(7 downto 0) <= ALU_r(7 downto 0);
    	          end if;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"18" |		-- CLC
                 x"38" |		-- SEC
                 x"58" |		-- CLI
                 x"78" |		-- SEI
                 x"B8" |		-- CLV
                 x"D8" |		-- CLD
                 x"F8" =>		-- SED
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;

                  case Opcode(7 downto 5) is
                    when "000" => P(Flag_C) <= '0';
                    when "001" => P(Flag_C) <= '1';
                    when "010" => P(Flag_I) <= '0';
                    when "011" => P(Flag_I) <= '1';
                    when "101" => P(Flag_V) <= '0';
                    when "110" => P(Flag_D) <= '0';
                    when "111" => P(Flag_D) <= '1';
                    when others => null;
                  end case;
                when "00100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"1A" |		-- INA
                 x"3A" |		-- DEA
                 x"88" |		-- DEY
                 x"C8" |		-- INY
                 x"CA" |		-- DEX
                 x"E8" =>		-- INX
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  Operation <= opTxx;

                  case Opcode is
                    when x"1A" | x"3A" =>
                      if (Opcode(5) = '1') then
                        Accu <= Accu - 1;
                        ALU_a <= Accu - 1;
                      else
                        Y <= Y + 1;
                        ALU_a <= Y + 1;
                      end if;
                    when x"88" | x"C8" =>
                      if (Opcode(6) = '0') then
                        Y <= Y - 1;
                        ALU_a <= Y - 1;
                      else
                        Y <= Y + 1;
                        ALU_a <= Y + 1;
                      end if;
                    when x"CA" | x"E8" =>
                      if (Opcode(5) = '0') then
                        X <= X - 1;
                        ALU_a <= X - 1;
                      else
                        X <= X + 1;
                        ALU_a <= X + 1;
                      end if;
                    when others => null;
                  end case;
                when "00100" =>
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;
                when "00110" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"1B" |		-- TCS / TAS
	         x"3B" |		-- TSC / TSA
	         x"5B" |		-- TCD / TAD
	         x"7B" |		-- TDC / TDA
	         x"8A" |		-- TXA
                 x"98" |		-- TYA
                 x"9A" |		-- TXS
                 x"A8" |		-- TAY
                 x"AA" |		-- TAX
                 x"BA" =>		-- TSX
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  Operation <= opTxx;

                  case Opcode is
                    when x"1B" =>
                      SP <= Accu;
                      ALU_a <= Accu;
                    when x"3B" =>
                      Accu <= SP;
                      ALU_a <= SP;
                    when x"5B" =>
                      DR <= Accu;
                      ALU_a <= Accu;
                    when x"7B" =>
                      Accu <= DR;
                      ALU_a <= DR;
                    when x"8A" =>
                      Accu(7 downto 0) <= X(7 downto 0);
                      ALU_a <= X;
                    when x"98" =>
                      Accu(7 downto 0) <= Y(7 downto 0);
                      ALU_a <= Y;
                    when x"9A" =>
                      SP(7 downto 0) <= X(7 downto 0);
                      ALU_a <= X;
                    when x"A8" =>
                      Y(7 downto 0) <= Accu(7 downto 0);
                      ALU_a <= Accu;
                    when x"AA" =>
                      X(7 downto 0) <= Accu(7 downto 0);
                      ALU_a <= Accu;
                    when x"BA" =>
                      X(7 downto 0) <= SP(7 downto 0);
                      ALU_a <= SP;
                    when others => null;
                  end case;
                when "00100" =>
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"20" =>		-- JSR
              case OpcCounter is
                when "00010" =>		-- read new address
                  PC <= PC + 1;
                when "00100" =>
                  PC <= PC + 1;
                  Tmp2(7 downto 0) <= IntData;		-- read LB address
                when "00110" =>		-- push HB PC to Stack
                  AddressMode <= amSP;
                  D_out <= PC(15 downto 8);
                  Tmp2(15 downto 8) <= IntData;		-- read HB address
                  RW_i <= '0';
                when "01000" =>		-- push LB PC to Stack
                  SP <= SP - 1;
                  D_out <= PC(7 downto 0);
                when "01010" =>
                  SP <= SP - 1;
                  AddressMode <= amPC;
                  PC <= Tmp2;
                  RW_i <= '1';
                when "01100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"22" =>		-- JSL
              case OpcCounter is
                when "00010" =>		-- read new address
                  PC <= PC + 1;
                when "00100" =>
                  PC <= PC + 1;
                  Tmp2(7 downto 0) <= IntData;	-- read LB address
                when "00110" =>
                  PC <= PC + 1;
                  Tmp2(7 downto 0) <= IntData;	-- read HB address
                when "01000" =>		-- push PBR to Stack
                  AddressMode <= amSP;
                  D_out <= PBR;
                  Tmp2(15 downto 8) <= IntData;	-- read PBR address
                  RW_i <= '0';
                when "01010" =>		-- push HB PC to Stack
                  SP <= SP - 1;
                  D_out <= PC(15 downto 8);
                when "01100" =>		-- push LB PC to Stack
                  SP <= SP - 1;
                  D_out <= PC(7 downto 0);
                when "01110" =>
                  SP <= SP - 1;
                  AddressMode <= amPC;
                  PC <= Tmp2;
                  RW_i <= '1';
                when "10000" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"24" |		-- BIT $xx
                 x"34" =>		-- BIT $xx,X
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;
                when "00100" =>		-- read zero page address
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;
                  Tmp <= DR + IntData;
                  Operation <= opBIT;
                  ALU_a <= Accu;

                  if (Opcode(4) = '0') then
                    Tmp <= DR + IntData;
                  else 
                    Tmp <= DR + (x"00" & (IntData + X(7 downto 0)));
                  end if;
                when "00110" =>		-- calculate and store results
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  PC <= PC + 1;
                  P(Flag_N) <= ALU_N;
                  P(Flag_V) <= ALU_V;
                  P(Flag_Z) <= ALU_Z;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"28" |		-- PLP
                 x"2B" |		-- PLD
                 x"68" |		-- PLA
                 x"7A" |		-- PLY
                 x"FA" =>		-- PLX
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  SP <= SP + 1;
                when "00100" =>		-- COMPATIBILITY
                  AddressMode <= amSP;
                  AddressMSB <= ahZero;
                  Operation <= opTxx;
                when "00110" =>
                  ALU_a(7 downto 0) <= Intdata;

                  case Opcode is
    	-- store read byte into Status register but set bit 5 !!!
                    when x"28" => P <= IntData or x"20";
    	-- store read byte in DBR / Accu / Y / X
                    when x"2B" => DBR <= IntData;
                    when x"68" => Accu(7 downto 0) <= IntData;
                    when x"7A" => Y(7 downto 0) <= IntData;
                    when x"FA" => X(7 downto 0) <= IntData;
                    when others => null;
                  end case;
                when "01000" =>
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  SP <= SP - 1;

		  if not (Opcode = x"28") then
                    P(Flag_N) <= ALU_N;
                    P(Flag_Z) <= ALU_Z;
                  end if;  

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"2C" |		-- BIT $xxxx
                 x"3C" =>		-- BIT $xxxx,X
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;
                when "00100" =>		-- read LB address
                  PC <= PC + 1;
                  Tmp2(7 downto 0) <= IntData;
                when "00110" =>		-- read HB address
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;

                  if (Opcode(4) = '0') then
                    Tmp <= IntData & Tmp2(7 downto 0);
		  else
                    Tmp <= (IntData & Tmp2(7 downto 0)) + X;
                  end if;

                  Operation <= opBIT;
                  ALU_a <= Accu;
                when "01000" =>		-- calculate and store results
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  PC <= PC + 1;
                  P(Flag_N) <= ALU_N;
                  P(Flag_V) <= ALU_V;
                  P(Flag_Z) <= ALU_Z;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"40" =>		-- RTI
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  SP <= SP + 1;
                when "00100" =>		-- COMPATIBILITY
                  AddressMode <= amSP;
                  AddressMSB <= ahZero;
                when "00110" =>
                  SP <= SP + 1;
                  P <= IntData;		-- store read byte into Status register
                when "01000" =>
                  SP <= SP + 1;
                  PC(7 downto 0) <= IntData;	-- store read byte into PC
                when "01010" =>
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  SP <= SP + 1;
                  PC(15 downto 8) <= IntData;	-- store read byte into PC
                when "01100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"42" =>		-- WDM, future use
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 2;
                when "00100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"44" |		-- MVP
                 x"54" =>		-- MVN
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
-- ###### to do                  
                when "00100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"4C" =>		-- JMP $xxxx
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                when "00100" =>
                  PC <= PC + 1;
                  Tmp2(7 downto 0) <= IntData;	-- read low byte new address
                when "00110" =>
                  PC <= IntData & Tmp2(7 downto 0);	-- read high byte new address
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"5C" =>		-- JML $xxxxxx
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                when "00100" =>
                  PC <= PC + 1;
                  Tmp(7 downto 0) <= IntData;	-- read low byte new address
                when "00110" =>
                  PC <= PC + 1;
                  Tmp(15 downto 8) <= IntData;	-- read high byte new address
                when "01000" =>
                  PBR <= IntData;	-- read PBR new address
                  PC <= TMP;
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"60" =>		-- RTS
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  SP <= SP + 1;
                when "00100" =>		-- COMPATIBILITY
                  AddressMode <= amSP;
                  AddressMSB <= ahZero;
                when "00110" =>
                  PC(7 downto 0) <= IntData;	-- store read byte into PC
-- shorter ???                  
-- ####                 Tmp(7 downto 0) <= IntData;	-- store read byte into PC
                  SP <= SP + 1;
                when "01000" =>
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  PC(15 downto 8) <= IntData;	-- store read byte into PC
-- ####                 PC <= (IntData & Tmp(7 downto 0)) + 1; -- store read byte into PC
                when "01010" =>
                  PC <= PC + 1;
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"62" =>		-- PER
              case OpcCounter is
                when "00010" =>		-- read LB offset
                  PC <= PC + 1;
                when "00100" =>		-- read HB offset
                  Tmp2(7 downto 0) <= IntData;
                  PC <= PC + 1;
                when "00110" =>
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;
                  Tmp <= PC + (Intdata & Tmp2(7 downto 0)) + 1;
                when "01000" =>
                  AddressMode <= amSP;
                  AddressMSB <= ahZero;
                  Tmp2(7 downto 0) <= IntData;



                  AddressMSB <= ahPBR;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"6B" =>		-- RTL
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  SP <= SP + 1;
                when "00100" =>		-- COMPATIBILITY
                  AddressMode <= amSP;
                  AddressMSB <= ahZero;
                when "00110" =>
                  PC(7 downto 0) <= IntData;	-- store read byte into PC
                  SP <= SP + 1;
                when "01000" =>
                  PC(15 downto 8) <= IntData;	-- store read byte into PC
                  SP <= SP + 1;
                when "01010" =>
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  PBR <= IntData;	-- store read byte into DPR
                when "01100" =>
                  PC <= PC + 1;
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"6C" =>		-- JMP ($xxxx)
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                when "00100" =>
                  PC <= PC + 1;
                  Tmp(7 downto 0) <= IntData;	-- read low byte indexed address
                when "00110" =>
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;
                  Tmp(15 downto 8) <= IntData;	-- read high byte indexed address
                when "01000" =>
                  Tmp <= Tmp+ 1;
                  PC(7 downto 0) <= IntData;	-- read low byte new address
                when "01010" =>
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  PC(15 downto 8) <= IntData;	-- read high byte new address

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"7C" =>		-- JMP ($xxxx,X)
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                when "00100" =>
                  PC <= PC + 1;
                  Tmp(7 downto 0) <= IntData;	-- read low byte indexed address
                when "00110" =>
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;
                  Tmp(15 downto 8) <= IntData;	-- read high byte indexed address
                when "01000" =>
                  Tmp <= Tmp + X;
                when "01010" =>
                  Tmp <= Tmp+ 1;
                  PC(7 downto 0) <= IntData;	-- read low byte new address
                when "01100" =>
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  PC(15 downto 8) <= IntData;	-- read high byte new address

                  OpcC_reset <= '1';
                when others => null;
              end case;
--  Idea:
--  Tmp <= X;
--  Tmp <= Tmp + Intdata;
--  Tmp <= Tmp + (Indata & x"00");              


            when x"82" =>		-- BRL
              case OpcCounter is
                when "00010" =>		-- read LB offset
                  PC <= PC + 1;
                when "00100" =>		-- read HB offset
                  Tmp2(7 downto 0) <= IntData;
                  PC <= PC + 1;
                when "00110" =>
                  PC <= PC + (Intdata & Tmp2(7 downto 0)) + 1;
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"84" |		-- STY $xx
                 x"85" |		-- STA $xx
                 x"86" | 		-- STX $xx

                 x"64" | 		-- STZ $xx
                 x"74" |		-- STZ $xx,X

                 x"94" |		-- STY $xx,X
                 x"95" |		-- STA $xx,X
                 x"96" =>		-- STX $xx,Y
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;
                when "00100" =>		-- read LB indexed address
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;
                  RW_i <= '0';

                  if (Opcode(4) = '0') then
                    Tmp <= DR + IntData;
                  elsif (Opcode(1) = '1') then
                    Tmp <= DR + (x"00" & (IntData + Y(7 downto 0)));
                  else
                    Tmp <= DR + (x"00" & (IntData + X(7 downto 0)));
                  end if;

                  case Opcode(1 downto 0) is
                    when "00" =>
                      if (Opcode(7) = '1') then   
                        D_out <= Y(7 downto 0);
                      else
		        D_out <= x"00";
                      end if;
                    when "01" => D_out <= Accu(7 downto 0);
                    when "10" => D_out <= X(7 downto 0);
                    when others => null;
                  end case;
                when "00110" =>
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  PC <= PC + 1;
                  RW_i <= '1';

                  if (Opcode(4) = '0') then
                    OpcC_reset <= '1';
                  end if;
                when "01000" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"89" =>		-- BIT #n
              case OpcCounter is
                when "00010" =>		-- read byte
                  PC <= PC + 1;
                  ALU_a <= Accu;
                  Operation <= opBIT;
                when "00100" =>		-- calculate and store results
                  PC <= PC + 1;
                  P(Flag_N) <= ALU_N;
                  P(Flag_V) <= ALU_V;
                  P(Flag_Z) <= ALU_Z;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"8C" |		-- STY $xxxx
                 x"8D" |		-- STA $xxxx
                 x"8E" | 		-- STX $xxxx

                 x"9C" | 		-- STZ $xxxx
                 x"9E" |		-- STZ $xxxx,X

                 x"99" |		-- STA $xxxx,Y
                 x"9D" =>		-- STA $xxxx,X
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;
                when "00100" =>		-- read LB indexed address
                  PC <= PC + 1;
                  Tmp2(7 downto 0) <= IntData;
                when "00110" =>		-- read HB address
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;
                  RW_i <= '0';

                  if (Opcode(4) = '0') then
                    Tmp <= IntData & Tmp2(7 downto 0);
                  elsif (Opcode(2) = '0') then
                    Tmp <= IntData & Tmp2(7 downto 0) + Y;
                  else
                    Tmp <= IntData & Tmp2(7 downto 0) + X;
                  end if;

                  case Opcode(1 downto 0) is
                    when "00" => 
                      if (Opcode(4) = '0') then   
                        D_out <= Y(7 downto 0);
                      else
		        D_out <= x"00";
                      end if;
                    when "01" => D_out <= Accu(7 downto 0);
                    when "10" => 
                      if (Opcode(4) = '0') then   
                        D_out <= Y(7 downto 0);
                      else
		        D_out <= x"00";
                      end if;
                    when others => null;
                  end case;
                when "01000" =>
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  PC <= PC + 1;
                  RW_i <= '1';

                  if (Opcode(4) = '0') then
                    OpcC_reset <= '1';
                  end if;  
                when "01010" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"91" |		-- STA ($xx),Y
                 x"92" =>		-- STA ($xx)
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;
                when "00100" =>		-- read zero page address
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;
                  Tmp <= DR + IntData;
                when "00110" =>		-- read HB indexed address
                  Tmp <= (Tmp + 1) and x"00FF";
                  Tmp2(7 downto 0) <= IntData;		-- store LB indexed address
                when "01000" =>		-- read byte at indexed address
                  if (Opcode(0) = '1') then
                    Tmp <= (IntData  & Tmp2(7 downto 0)) + Y;
                  else
                    Tmp <= IntData  & Tmp2(7 downto 0);
                  end if;
		    
                  D_out <= Accu(7 downto 0);
                  RW_i <= '0';
                when "01010" =>		-- calculate and store results
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  PC <= PC + 1;
                  RW_i <= '1';
                when "01100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"A0" |		-- LDY #nn
                 x"A2" |		-- LDX #nn
                 x"A9" =>		-- LDA #nn
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  Operation <= opLDx;
                when "00100" =>
                  case Opcode(2 downto 0) is
                    when "000" => Y(7 downto 0) <= IntData;
                    when "001" => Accu(7 downto 0) <= IntData;
                    when "010" => X(7 downto 0) <= IntData;
                    when others => null;
                  end case;

                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;
                  PC <= PC + 1;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"AC" |		-- LDY $xxxx
                 x"AE" |		-- LDX $xxxx
                 x"BC" |		-- LDY $xxxx,X
                 x"BE" =>		-- LDX $xxxx,Y
              case OpcCounter is
                when "00010" =>		-- read zero page address
                  PC <= PC + 1;
                  Operation <= opLDx;
                when "00100" =>		-- read LB address
                  Tmp2(7 downto 0) <= IntData;
                  PC <= PC + 1;
                  
                when "00110" =>		-- read HB address
                  ALU_a <= Accu;
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;

                  if (Opcode(4) = '0') then
                    Tmp <= IntData & Tmp2(7 downto 0);
		  elsif (Opcode(1) = '0') then
                    Tmp <= (IntData & Tmp2(7 downto 0)) + X;
		  else
                    Tmp <= (IntData & Tmp2(7 downto 0)) + Y;
                  end if;

                  if (M16_X = '0') then
                    DoubleAction <= '1';
                    Operation <= '1' & Operation(4 downto 0);
                  end if;
                when "01000" =>		-- calculate and store results
                  if (DoubleAction = '1') then
                    TmpAlu <= Intdata;
                    OpcCounter := OpcCounter - 2;	-- "00110"
                    DoubleAction <= '0';
                    Tmp <= Tmp + 1;
                  else  
                    AddressMode <= amPC;
                    AddressMSB <= ahPBR;
                    PC <= PC + 1;
                    P(Flag_N) <= ALU_N;
                    P(Flag_Z) <= ALU_Z;

                    case Opcode(6 downto 5) is
                      when "11" =>
                        P(Flag_V) <= ALU_V;
                        P(Flag_C) <= ALU_C;
                      when others => null;
                    end case;

                    if (Opcode(1) = '0') then
                      Y <= ALU_r;
                    else  
                      X <= ALU_r;
                    end if;

                    OpcC_reset <= '1';
                  end if;
                when others => null;
              end case;


            when x"C0" |		-- CPY #n
                 x"C9" |		-- CMP #n
                 x"E0" =>		-- CPX #n
              case OpcCounter is
                when "00010" =>		-- read byte
                  PC <= PC + 1;
                  Operation <= opCMP;

                  case Opcode(5 downto 3) is
                    when "000" => ALU_a <= Y;
                    when "001" => ALU_a <= Accu;
                    when "100" => ALU_a <= X;
                    when others => null;		
                  end case;

                  if (Opcode(0) = '1') then
                    if (M16_A = '0') then
                      DoubleAction <= '1';
                    end if;
                  else
                    if (M16_X = '0') then
                      DoubleAction <= '1';
                    end if;
                  end if;
                when "00100" =>		-- calculate and store results
                  if (DoubleAction = '1') then
		    TmpAlu <= Intdata;
                    OpcCounter := OpcCounter - 2;	-- "00010"
                    DoubleAction <= '0';
                  end if; 

                  PC <= PC + 1;
                  P(Flag_C) <= ALU_C;
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"C4" |		-- CPY $xx
                 x"C5" |		-- CMP $xx
                 x"D5" |		-- CMP $xx,X
                 x"E4" =>		-- CPX $xx
              case OpcCounter is
                when "00010" =>		-- read byte
                  PC <= PC + 1;
                  Operation <= opCMP;

                  case Opcode(5 downto 0) is
                    when "000100" => ALU_a <= Y;
                    when "100100" => ALU_a <= X;
                    when others => ALU_a <= Accu;
                  end case;
                when "00100" =>		-- calculate and store results
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;

                  if (Opcode = x"D5") then
                    Tmp <= DR + (x"00" & (IntData + X(7 downto 0)));
                  else
                    Tmp <= DR + IntData;
                  end if;
                when "00110" =>		-- calculate and store results
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  PC <= PC + 1;
                  P(Flag_C) <= ALU_C;
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"CB" =>		-- WAI
              case OpcCounter is
                when "00010" =>
                  WAI_flg <= '1';
                when "00100" =>
                  if (WAI_flg = '0') then
                    OpcC_reset <= '1';
                  end if;
                when "00110" =>
                  OpcCounter := "00100";
                when others => null;
              end case;


            when x"CC" |		-- CPY $xxxx
                 x"CD" |		-- CMP $xxxx
                 x"EC" |		-- CPX $xxxx

                 x"D9" |		-- CMP $xxxx,Y
                 x"DD" =>		-- CMP $xxxx,X
              case OpcCounter is
                when "00010" =>			-- read byte
                  PC <= PC + 1;
                  Operation <= opCMP;

                  case Opcode is
                    when x"CC" => ALU_a <= Y;
                    when x"EC" => ALU_a <= X;
                    when others => ALU_a <= Accu;
                  end case;
                when "00100" =>			-- read LB address
                  PC <= PC + 1;
                  Tmp2(7 downto 0) <= IntData;
                when "00110" =>		-- read HB address
                  AddressMode <= amTmp;
                  AddressMSB <= ahDBR;

                  case Opcode(4 downto 2) is
		    when "111"  => Tmp <= (IntData & Tmp2(7 downto 0)) + X;
                    when "110"  => Tmp <= (IntData & Tmp2(7 downto 0)) + Y;
                    when others => Tmp <= IntData & Tmp2(7 downto 0);
                  end case;
                when "01000" =>			-- calculate and store results
                  AddressMode <= amPC;
                  AddressMSB <= ahPBR;
                  PC <= PC + 1;
                  P(Flag_C) <= ALU_C;
                  P(Flag_N) <= ALU_N;
                  P(Flag_Z) <= ALU_Z;

                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"DB" =>		-- STP
              OpcCounter := "00010";


            when x"EA" =>		-- NOP
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                when "00100" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"EB" =>		-- XBA
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  Tmp <= Accu;
                when "00100" =>
                  Accu(15 downto 8) <= Tmp(7 downto 0);
                  Accu(7 downto 0) <= Tmp(15 downto 8);
                when "00110" =>		-- COMPATIBILITY
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when x"FB" =>		-- XCE
              case OpcCounter is
                when "00010" =>
                  PC <= PC + 1;
                  Pn(Flag_Emu) <= P(Flag_C);
                when "00100" =>		-- COMPATIBILITY
                  if (P(Flag_C) = '1') then
                    X(7 downto 0) <= x"00";
                    Y(7 downto 0) <= x"00";
                  end if;
                  
                  OpcC_reset <= '1';
                when others => null;
              end case;


            when others => null;
          end case;
        end if;
      end if;
    end if;
  end process;



--  ====  ALU  =====

-- Select use of decimal flag
  with Operation select ALU_D <=
    '0'		when opBRA | opCMP | opwBRA | opwCMP,
    P(Flag_D)	when others;

-- Select use of carry flag
  with Operation select ALU_Ci <=
    '0'		when opBRA | opwBRA ,
    '1'		when opCMP | opwCMP ,
    P(Flag_C)	when others;


-- Set second operand
  with Operation select ALU_b <=
    x"00" & not IntData		when opSBC | opCMP,
    not (IntData & TmpAlu)	when opwSBC | opwCMP,
    ALU_a			when opASLa | opLSRa | opROLa | opRORa |
                                     opwASLa | opwLSRa | opwROLa | opwRORa,
    PC				when opBRA | opwBRA,
    x"00" & IntData		when opAND | opOR  | opEOR | opASL | opLSR |
                                     opROL | opROR | opINC | opDEC | opADC |
                                     opBIT | opLDx | opTxx | opTRB | opTSB,
    IntData & TmpAlu		when others;


-- Add 1st and 2nd operand
  NibbleL <=('0' & ALU_a(3 downto 0)) + ('0' & ALU_b(3 downto 0))
              + ("0000" & ALU_Ci);

--  Gideon    if (   (NibbleL(4) = '1') or (NibbleL(3 downto 2) = "11")
--	        or (NibbleL(3 downto 1) = "101")) then
  NibbleL2 <= NibbleL + 6 when ((ALU_D = '1') and (NibbleL > 9)) else NibbleL;

  NibbleH <= ('0' & ALU_a(7 downto 4)) + ('0' & ALU_b(7 downto 4))
               + ("0000" & NibbleL2(4));

--  Gideon    if (   (NibbleH(4) = '1') or (NibbleH(3 downto 2) = "11")
--	        or (NibbleH(3 downto 1) = "101")
--		or ((NibbleH(4) = '1') and (NibbleH(3 downto 0) = "1001"))) then -- >9 (10-11, 12-15)
  NibbleH2 <= NibbleH + 6 when ((ALU_D = '1') and (NibbleH > 9)) else NibbleH;

  NibbleWL <= ('0' & ALU_a(11 downto 8)) + ('0' & ALU_b(11 downto 8))
               + ("0000" & NibbleH2(4));

  NibbleWL2 <= NibbleWL + 6 when ((ALU_D = '1') and (NibbleWL > 9)) else NibbleWL;

  NibbleWH <= ('0' & ALU_a(15 downto 12)) + ('0' & ALU_b(15 downto 12))
               + ("0000" & NibbleWL2(4));

  NibbleWH2 <= NibbleWH + 6 when ((ALU_D = '1') and (NibbleWH > 9)) else NibbleWH;


-- Calculate result
  with Operation select ALU_r <=
    NibbleWH2(3 downto 0) & NibbleWL2(3 downto 0) &
    NibbleH2(3 downto 0) & NibbleL2(3 downto 0)		when opADC | opSBC | opCMP |
                                                             opBRA | opwADC | opwSBC |
							     opwCMP | opwBRA,
    ALU_a and ALU_b					when opAND | opBIT | 
                                                             opwAND | opwBIT,
    ALU_b(14 downto 0) & "0"				when opASL | opASLa |
                                                             opwASL | opwASLa,
    ALU_b - 1						when opDEC | opwDEC,
    ALU_a xor ALU_b					when opEOR | opwEOR,
    ALU_b + 1						when opINC | opwINC,
    x"00" & "0" & ALU_b(7 downto 1)			when opLSR | opLSRa,
    "0" & ALU_b(15 downto 1)				when opwLSR | opwLSRa,
    ALU_a or ALU_b					when opOR | opwOR,
    ALU_b(14 downto 0) & ALU_Ci				when opROL | opROLa |
                                                             opwROL | opwROLa,
    x"00" & ALU_Ci & ALU_b(7 downto 1)			when opROR | opRORa,
    ALU_Ci & ALU_b(15 downto 1)				when opwROR | opwRORa,
    (ALU_a xor x"FFFF") and ALU_b			when opTRB | opwTRB,
    ALU_a or ALU_b					when opTSB | opwTSB,
    ALU_a						when opTxx | opwTxx,
    ALU_b						when others;


-- Calculate Carry flag
  with Operation select ALU_C <=
    ALU_b(7)		when opASL | opROL | opASLa | opROLa,
    ALU_b(15)		when opwASL | opwROL | opwASLa | opwROLa,
    ALU_b(0)		when opLSR | opROR | opLSRa | opRORa |
                             opwLSR | opwROR | opwLSRa | opwRORa,
    NibbleH2(4)		when opADC | opBRA | opCMP | opSBC,
    NibbleWH2(4)	when others;	


-- Calculate Negative flag
  with Operation select ALU_N <=
    IntData(7)	when opBIT | opwBIT,
    ALU_r(7)	when opAND | opOR  | opEOR | opASL | opLSR | opROL | opROR |
                     opINC | opDEC | opADC | opLDx | opTxx | opTRB | opTSB | 
		     opCMP | opSBC | opASLa | opLSRa | opROLa | opRORa,
    ALU_r(15)	when others;


-- Calculate Overflow flag
  with Operation select ALU_V <=
    ((ALU_a(7) xor ALU_r(7)) and (ALU_b(7) xor ALU_r(7)))
           	when opADC | opSBC | opCMP | opBRA,
    ((ALU_a(15) xor ALU_r(15)) and (ALU_b(15) xor ALU_r(15)))
           	when opwADC | opwSBC | opwCMP | opwBRA,
    IntData(6)	when others;


-- Calculate Zero flag
  ALU_Z <= '1'	when (   ((ALU_r = 0) and (Operation(5) = '1'))
                      or ((ALU_r(7 downto 0) = 0) and (Operation(5) = '0')))
                else '0';


-- Select whether Branch should be taken or not
  with Opcode(7 downto 4) select Branch <=
    not P(Flag_N)	when "0001",		-- BPL
    P(Flag_N) 		when "0011",		-- BMI
    not P(Flag_V) 	when "0101",		-- BVC
    P(Flag_V) 		when "0111",		-- BVS
    not P(Flag_C) 	when "1001",		-- BCC
    P(Flag_C) 		when "1011",		-- BCS
    not P(Flag_Z) 	when "1101",		-- BNE
    P(Flag_Z)	 	when "1111",		-- DEQ
    '1'			when others;		-- BRA

end AR_RB65816;

